import os
import pandas as pd
import numpy as np
import copy
import logging
from tqdm import tqdm

from py_expression_eval import Parser

from massql import msql_parser
from massql import msql_fileloading
from massql import msql_engine_filters
from massql.msql_engine_filters import _get_mz_tolerance, _get_minintensity

math_parser = Parser()
console = logging.StreamHandler()
console.setLevel(logging.INFO)

def DEBUG_MSG(msg):
    import sys

    print(msg, file=sys.stderr, flush=True)

def init_ray():
    import ray
    if not ray.is_initialized():
        ray.init(ignore_reinit_error=True, object_store_memory=8000000000, num_cpus=8)


def _get_ppm_tolerance(qualifiers):
    if qualifiers is None:
        return None

    if "qualifierppmtolerance" in qualifiers:
        ppm = qualifiers["qualifierppmtolerance"]["value"]
        return ppm

    return None

def _get_da_tolerance(qualifiers):
    if qualifiers is None:
        return None

    if "qualifiermztolerance" in qualifiers:
        return qualifiers["qualifiermztolerance"]["value"]

    return None


def process_query(input_query, input_filename, path_to_grammar=None, cache=True, parallel=False, ms1_df=None, ms2_df=None):
    """
    Process an actual query

    Args:
        input_query ([type]): [description]
        input_filename ([type]): [description]
        path_to_grammar ([type], optional): [description]. Defaults to None.
        cache (bool, optional): [description]. Defaults to True.
        parallel (bool, optional): [description]. Defaults to False.
        ms1_df ([type], optional): [description]. Defaults to None. Pass in if you have these data in memory
        ms2_df ([type], optional): [description]. Defaults to None. Pass in if you have these data in memory

    Returns:
        query results data frame: [description]
    """

    parsed_dict = msql_parser.parse_msql(input_query, path_to_grammar=path_to_grammar)

    return _evalute_variable_query(parsed_dict, input_filename, cache=cache, parallel=parallel, ms1_df=ms1_df, ms2_df=ms2_df)

def _determine_mz_max(mz, ppm_tol, da_tol):
    da_tol = da_tol if da_tol < 10000 else 0
    ppm_tol = ppm_tol if ppm_tol < 10000 else 0
    # We are going to make the bins half of the actual tolerance
    half_delta = max(mz * ppm_tol / 1000000, da_tol) / 2

    half_delta = half_delta if half_delta > 0 else 0.05

    return mz + half_delta

def _evalute_variable_query(parsed_dict, input_filename, cache=True, parallel=False, ms1_df=None, ms2_df=None):
    # Lets check if there is a variable in here, the only one allowed is X
    for condition in parsed_dict["conditions"]:
        try:
            if "querytype" in condition["value"][0]:
                subquery_val_df = _evalute_variable_query(
                    condition["value"][0], input_filename, cache=cache
                )
                condition["value"] = list(
                    subquery_val_df["precmz"]
                )  # Flattening results
        except:
            pass

    # Variable Expression Parameters
    variable_properties = {}
    variable_properties["has_variable"] = False
    variable_properties["ppm_tolerance"] = 100000
    variable_properties["da_tolerance"] = 100000
    variable_properties["query_ms1"] = False
    variable_properties["query_ms2"] = False
    variable_properties["query_ms2prec"] = False
    variable_properties["min"] = 0
    variable_properties["max"] = 1000000
    variable_properties["mindefect"] = 0
    variable_properties["maxdefect"] = 1
    

    # Checking for ranges in query
    new_conditions = []
    for condition in parsed_dict["conditions"]:
        if condition["type"] == "xcondition":
            if "min" in condition:
                variable_properties["min"] = condition["min"]
                variable_properties["max"] = condition["max"]
            if "mindefect" in condition:
                variable_properties["mindefect"] = condition["mindefect"]
                variable_properties["maxdefect"] = condition["maxdefect"]
        else:
            new_conditions.append(condition)
    parsed_dict["conditions"] = new_conditions

    # Here we will check if there is a variable in the expression
    for condition in parsed_dict["conditions"]:
        if not "value" in condition:
            # If there is no value in the condition, e.g. retention time and mobility
            continue
        
        for value in condition["value"]:
            try:
                # Checking if X is in any string
                # TODO: Check multiple conditions, to see if any of them have an X, e.g. (101 OR X-10)
                if "X" in value[0]:
                    if value == "X":
                        # This is the main varaible, not expression containing it
                        if condition["type"] == "ms1mzcondition":
                            variable_properties["query_ms1"] = True
                        if condition["type"] == "ms2productcondition":
                            variable_properties["query_ms2"] = True
                        if condition["type"] == "ms2neutrallosscondition":
                            variable_properties["query_ms2"] = True
                        if condition["type"] == "ms2precursorcondition":
                            variable_properties["query_ms2prec"] = True

                    variable_properties["has_variable"] = True
                    #mz_tolerance = _get_mz_tolerance(condition.get("qualifiers", None), 1000000)

                    ppm_tolerance = _get_ppm_tolerance(condition.get("qualifiers", None))
                    da_tolerance = _get_da_tolerance(condition.get("qualifiers", None))

                    if da_tolerance is not None:
                        variable_properties["da_tolerance"] = min(variable_properties["da_tolerance"], da_tolerance)
                    if ppm_tolerance is not None:
                        variable_properties["ppm_tolerance"] = min(variable_properties["ppm_tolerance"], ppm_tolerance)                    
                    continue
            except TypeError:
                # This is when the target is actually a float
                pass

    # Loading data if not passed in 
    if ms1_df is None:
        ms1_df, ms2_df = msql_fileloading.load_data(input_filename, cache=cache)

    # Here we are going to translate the variable query into a concrete query based upon the data
    all_concrete_queries = []
    if variable_properties["has_variable"]:
        # Here we could do a pre-query without any of the other conditions
        presearch_parse = copy.deepcopy(parsed_dict)
        non_variable_conditions = []
        for condition in presearch_parse["conditions"]:
            if "value" in condition:
                for value in condition["value"]:
                    try:
                        # Checking if X is in any string
                        if "X" in value:
                            continue
                    except TypeError:
                        # This is when the target is actually a float
                        pass
                    non_variable_conditions.append(condition)
            elif "min" in condition:
                min_val = condition["min"]
                max_val = condition["max"]
                try:
                    if "X" in min_val:
                        continue
                except TypeError:
                    # This is when the target is actually a float
                    pass

                try:
                    if "X" in max_val:
                        continue
                except TypeError:
                    # This is when the target is actually a float
                    pass

                non_variable_conditions.append(condition)

        presearch_parse["conditions"] = non_variable_conditions

        ms1_df, ms2_df = _executeconditions_query(presearch_parse, input_filename, cache=cache)
        variable_x_ms1_df = ms1_df

        # Here we are trying to pre-filter conditions based upon the qualifiers to make the variable search space smaller
        for condition in parsed_dict["conditions"]:
            if not condition["conditiontype"] == "where":
                continue

            if "value" in condition:
                if not "X" in condition["value"]:
                    continue
        
            # Filtering MS1 peaks only to consider contention for X
            if condition["type"] == "ms1mzcondition":
                min_int, min_intpercent, min_tic_percent_intensity = _get_minintensity(condition.get("qualifiers", None))
                variable_x_ms1_df = ms1_df[
                    (ms1_df["i"] > min_int) & 
                    (ms1_df["i_norm"] > min_intpercent) & 
                    (ms1_df["i_tic_norm"] > min_tic_percent_intensity)]

            # TODO: Do this for other types of variables

        # Here we will start with the smallest mass and then go up
        masses_considered_df_list = []
        if variable_properties["query_ms1"]:
            masses_considered_df_list.append(variable_x_ms1_df["mz"])
        if variable_properties["query_ms2"]:
            masses_considered_df_list.append(ms2_df["mz"])
        if variable_properties["query_ms2prec"]:
            masses_considered_df_list.append(ms2_df["precmz"])

        masses_considered_df = pd.DataFrame()
        masses_considered_df["mz"] = pd.concat(masses_considered_df_list)
        # NOTE: This might cause bugs, we might consider every mass within every single scan, or at least we could make the tolernace as even smaller than half the max
        masses_considered_df["mz_max"] = masses_considered_df["mz"].apply(lambda x: _determine_mz_max(x, variable_properties["ppm_tolerance"], variable_properties["da_tolerance"]))
        
        masses_considered_df = masses_considered_df.sort_values("mz")
        masses_list = masses_considered_df.to_dict(orient="records")

        running_max_mz = 0
        for masses_obj in tqdm(masses_list):
            mz_val = masses_obj["mz"]

            if running_max_mz > mz_val:
                continue

            # Cheking the validity of the mz_val
            if mz_val < variable_properties["min"] or mz_val > variable_properties["max"]:
                continue
            mz_val_defect = mz_val - int(mz_val)
            if mz_val_defect < variable_properties["mindefect"] or mz_val_defect > variable_properties["maxdefect"]:
                continue

            #######################
            # Writing new query
            #######################
            substituted_parse = copy.deepcopy(parsed_dict)
            
            for condition in substituted_parse["conditions"]:
                # This is for standard conditions
                if "value" in condition:
                    for i, value in enumerate(condition["value"]):
                        # Rewriting the condition value
                        try:
                            if "X" in value:
                                new_value = math_parser.parse(value).evaluate({
                                    "X" : mz_val
                                })
                                condition["value"][i] = new_value
                        except TypeError:
                            # This is when the target is actually a float
                            pass

                        # Rewriting the qualifier values
                        try:
                            if "qualifiers" in condition:
                                for qualifier in condition["qualifiers"]:
                                    if "qualifier" in qualifier:
                                        if "value" in condition["qualifiers"][qualifier]:
                                            old_value = condition["qualifiers"][qualifier]["value"]
                                            condition["qualifiers"][qualifier]["value"] = old_value.replace("X", str(mz_val))
                        except AttributeError:
                            pass
                
                # TODO: For other types of conditions that might include variables
                if "min" in condition:
                    # Rewriting the condition min
                    value = condition["min"]
                    try:
                        if "X" in value:
                            new_value = math_parser.parse(value).evaluate({
                                "X" : mz_val
                            })
                            condition["min"] = new_value
                    except TypeError:
                        # This is when the target is actually a float
                        pass

                if "max" in condition:
                    # Rewriting the condition min
                    value = condition["max"]
                    try:
                        if "X" in value:
                            new_value = math_parser.parse(value).evaluate({
                                "X" : mz_val
                            })
                            condition["max"] = new_value
                    except TypeError:
                        # This is when the target is actually a float
                        pass
            
            # Let's consider this mz
            running_max_mz = masses_obj["mz_max"]

            # Checking the x conditions
            substituted_parse["comment"] = str(mz_val)

            all_concrete_queries.append(substituted_parse)
    else:
        all_concrete_queries.append(parsed_dict)

    print("TOTAL QUERIES", len(all_concrete_queries))

    # Perfoming all the concrete queries
    collated_list = [] # This list holds the collated set of results from each query, final result is a concat of all of them

    # Ray Parallel Version
    execute_serial = True
    if parallel:
        import ray
        import msql_engine_ray

        if ray.is_initialized():
            # TODO: Divide up the parallel thing
            chunk_size = 100
            concrete_query_lists = [all_concrete_queries[i:i + chunk_size] for i in range(0, len(all_concrete_queries), chunk_size)]
            futures = [msql_engine_ray._executeconditions_query_ray.remote(concrete_query_list, input_filename, ms1_input_df=ms1_df, ms2_input_df=ms2_df, cache=cache) for concrete_query_list in concrete_query_lists]
            all_ray_results = ray.get(futures)

            # Flattening this list of lists
            collated_list = [item for sublist in all_ray_results for item in sublist]

            execute_serial = False
    
    # This is the fallback
    if execute_serial:
        # Serial Version
        for concrete_query in tqdm(all_concrete_queries):
            results_ms1_df, results_ms2_df = _executeconditions_query(concrete_query, input_filename, ms1_input_df=ms1_df, ms2_input_df=ms2_df, cache=cache)
            
            collated_df = _executecollate_query(parsed_dict, results_ms1_df, results_ms2_df)
            collated_list.append(collated_df)

    # Concatenating all the results
    collated_df = pd.concat(collated_list)
    collated_df = collated_df.reset_index(drop=True)

    # Lets try to remove duplicates
    try:
        if "comment" in collated_df:
            collated_df["truncated_comment"] = collated_df["comment"].astype(float).astype(int)
            collated_df = collated_df.drop_duplicates(subset=["scan", "truncated_comment"])
            collated_df = collated_df.drop("truncated_comment", axis=1)
    except:
        pass
        
    return collated_df


def _executeconditions_query(parsed_dict, input_filename, ms1_input_df=None, ms2_input_df=None, cache=True):
    # This function attempts to find the data that the query specifies in the conditions
    
    #import json
    #print("parsed_dict", json.dumps(parsed_dict, indent=4))

    # Let's apply this to real data
    if ms1_input_df is None and ms2_input_df is None:
        ms1_df, ms2_df = msql_fileloading.load_data(input_filename, cache=cache)
    else:
        ms1_df = ms1_input_df
        ms2_df = ms2_input_df

    # In order to handle intensities, we will make sure to sort all conditions with 
    # with the conditions that are the reference intensity first, then subsequent conditions
    # that have an intensity match will reference the saved reference intensities
    reference_conditions_register = {} # This will hold all the reference intensity values
    
    # This helps sort the qualifiers
    reference_conditions = []
    nonreference_conditions = []
    for condition in parsed_dict["conditions"]:
        if "qualifiers" in condition:
            if "qualifierintensityreference" in condition["qualifiers"]:
                reference_conditions.append(condition)
                continue
        nonreference_conditions.append(condition)
    all_conditions = reference_conditions + nonreference_conditions

    # These are for the WHERE clause, first lets filter by RT and polarity and scan
    for condition in all_conditions:
        if not condition["conditiontype"] == "where":
            continue

        # RT Filters
        if condition["type"] == "rtmincondition":
            rt = condition["value"][0]
            ms2_df = ms2_df[ms2_df["rt"] > rt]
            ms1_df = ms1_df[ms1_df["rt"] > rt]

            continue

        if condition["type"] == "rtmaxcondition":
            rt = condition["value"][0]
            ms2_df = ms2_df[ms2_df["rt"] < rt]
            ms1_df = ms1_df[ms1_df["rt"] < rt]

            continue
    
        # Polarity Filters
        if condition["type"] == "polaritycondition":
            polaritycondition = condition["value"][0]
            if polaritycondition == "positivepolarity":
                ms2_df = ms2_df[ms2_df["polarity"] == 1]
                ms1_df = ms1_df[ms1_df["polarity"] == 1]
            if polaritycondition == "negativepolarity":
                ms2_df = ms2_df[ms2_df["polarity"] == 2]
                ms1_df = ms1_df[ms1_df["polarity"] == 2]

            continue

        # Scan Filters
        if condition["type"] == "scanmincondition":
            scan = int(condition["value"][0])
            ms2_df = ms2_df[ms2_df["scan"] >= scan]
            ms1_df = ms1_df[ms1_df["scan"] >= scan]

            continue
            
        if condition["type"] == "scanmaxcondition":
            scan = int(condition["value"][0])
            ms2_df = ms2_df[ms2_df["scan"] <= scan]
            ms1_df = ms1_df[ms1_df["scan"] <= scan]

            continue

        # Charge Filters
        if condition["type"] == "chargecondition":
            charge = int(condition["value"][0])
            ms2_df = ms2_df[ms2_df["charge"] == charge]

            # Filtering the MS1 data now
            ms1_scans = set(ms2_df["ms1scan"])
            ms1_df = ms1_df[ms1_df["scan"].isin(ms1_scans)]

        # Mobility Filters
        if condition["type"] == "mobilitycondition":
            min_mobility = condition["min"]
            max_mobility = condition["max"]

            if "mobility" in ms2_df:
                ms2_df = ms2_df[(ms2_df["mobility"] >= min_mobility) & (ms2_df["mobility"] <= max_mobility)]

            if "mobility" in ms1_df.columns:
                ms1_df = ms1_df[(ms1_df["mobility"] >= min_mobility) & (ms1_df["mobility"] <= max_mobility)]

    # These are for the WHERE clause for peaks
    for condition in all_conditions:
        if not condition["conditiontype"] == "where":
            continue

        # Filtering MS2 Product Ions
        if condition["type"] == "ms2productcondition":
            ms1_df, ms2_df = msql_engine_filters.ms2prod_condition(condition, ms1_df, ms2_df, reference_conditions_register)
            continue

        # Filtering MS2 Precursor m/z
        if condition["type"] == "ms2precursorcondition":
            ms1_df, ms2_df = msql_engine_filters.ms2prec_condition(condition, ms1_df, ms2_df, reference_conditions_register)
            continue

        # Filtering MS2 Neutral Loss
        if condition["type"] == "ms2neutrallosscondition":
            ms1_df, ms2_df = msql_engine_filters.ms2nl_condition(condition, ms1_df, ms2_df, reference_conditions_register)
            continue

        # finding MS1 peaks
        if condition["type"] == "ms1mzcondition":
            ms1_df, ms2_df = msql_engine_filters.ms1_condition(condition, ms1_df, ms2_df, reference_conditions_register)
            continue

        skip_conditions = [ "rtmincondition", 
                            "rtmaxcondition", 
                            "polaritycondition", 
                            "scanmincondition", 
                            "scanmaxcondition",
                            "chargecondition",
                            "mobilitycondition"]
        
        if condition["type"] in skip_conditions:
            continue

        raise Exception("CONDITION NOT HANDLED")

    # These are for the FILTER clause
    for condition in all_conditions:
        if not condition["conditiontype"] == "filter":
            continue

        # filtering MS1 peaks
        if condition["type"] == "ms1mzcondition":
            if len(ms1_df) == 0:
                continue

            mz = condition["value"][0]
            mz_tol = 0.1
            mz_min = mz - mz_tol
            mz_max = mz + mz_tol
            ms1_df = ms1_df[(ms1_df["mz"] > mz_min) & (ms1_df["mz"] < mz_max)]

            continue
        
        if condition["type"] == "ms2productcondition":
            if len(ms2_df) == 0:
                continue

            mz = condition["value"][0]
            mz_tol = _get_mz_tolerance(condition.get("qualifiers", None), mz)
            mz_min = mz - mz_tol
            mz_max = mz + mz_tol

            min_int, min_intpercent, min_tic_percent_intensity = _get_minintensity(condition.get("qualifiers", None))

            ms2_df = ms2_df[(ms2_df["mz"] > mz_min) & 
                            (ms2_df["mz"] < mz_max) & 
                            (ms2_df["i"] > min_int) & 
                            (ms2_df["i_norm"] > min_intpercent) & 
                            (ms2_df["i_tic_norm"] > min_tic_percent_intensity)]

    if "comment" in parsed_dict:
        ms1_df["comment"] = parsed_dict["comment"]
        ms2_df["comment"] = parsed_dict["comment"]

    return ms1_df, ms2_df

def _executecollate_query(parsed_dict, ms1_df, ms2_df):
    # This function takes the dataframes from executing the conditions and returns the proper formatted version

    # Early Exit
    if len(ms1_df) == 0 and len(ms2_df) == 0:
        return pd.DataFrame()

    # collating the results
    if parsed_dict["querytype"]["function"] is None:
        if parsed_dict["querytype"]["datatype"] == "datams1data":
            return ms1_df
        if parsed_dict["querytype"]["datatype"] == "datams2data":
            return ms2_df
    else:
        # Applying function
        if parsed_dict["querytype"]["function"] == "functionscansum":

            # TODO: Fix how this scan is done so the result values for most things actually make sense
            if parsed_dict["querytype"]["datatype"] == "datams1data":
                if len(ms1_df) == 0:
                    return ms1_df

                ms1sum_df = ms1_df.groupby("scan").sum().reset_index()

                ms1_df = ms1_df.groupby("scan").first().reset_index()
                ms1_df["i"] = ms1sum_df["i"]

                return ms1_df
            if parsed_dict["querytype"]["datatype"] == "datams2data":
                if len(ms2_df) == 0:
                    return ms2_df

                ms2sum_df = ms2_df.groupby("scan").sum().reset_index()
                
                ms2_df = ms2_df.groupby("scan").first().reset_index()
                ms2_df["i"] = ms2sum_df["i"]

                return ms2_df

        if parsed_dict["querytype"]["function"] == "functionscanmz":
            if len(ms2_df) == 0:
                return ms2_df

            result_df = pd.DataFrame()
            result_df["precmz"] = list(set(ms2_df["precmz"]))
            return result_df

        if parsed_dict["querytype"]["function"] == "functionscannum":
            result_df = pd.DataFrame()

            if parsed_dict["querytype"]["datatype"] == "datams1data":
                result_df["scan"] = list(set(ms1_df["scan"]))
            if parsed_dict["querytype"]["datatype"] == "datams2data":
                result_df["scan"] = list(set(ms2_df["scan"]))

            return result_df

        # scaninfo return function
        if parsed_dict["querytype"]["function"] == "functionscaninfo":
            result_df = pd.DataFrame()

            if parsed_dict["querytype"]["datatype"] == "datams1data":
                if len(ms1_df) == 0:
                    return pd.DataFrame()

                groupby_columns = ["scan"]
                kept_columns = ["scan", "rt"]

                if "comment" in ms1_df:
                    groupby_columns.append("comment")
                    kept_columns.append("comment")

                if "mobility" in ms1_df:
                    kept_columns.append("mobility")

                result_df = ms1_df.groupby(groupby_columns).first().reset_index()
                result_df = result_df[kept_columns]
                result_df["mslevel"] = 1

                ms1sum_df = ms1_df.groupby(groupby_columns).sum().reset_index()
                ms1norm_df = ms1_df.groupby(groupby_columns).max().reset_index()
                result_df["i"] = ms1sum_df["i"]
                result_df["i_norm"] = ms1norm_df["i_norm"]
            if parsed_dict["querytype"]["datatype"] == "datams2data":
                if len(ms2_df) == 0:
                    return pd.DataFrame()

                kept_columns = ["scan", "precmz", "ms1scan", "rt", "charge"]
                groupby_columns = ["scan"]

                if "comment" in ms2_df:
                    groupby_columns.append("comment")
                    kept_columns.append("comment")

                if "mobility" in ms2_df:
                    kept_columns.append("mobility")

                result_df = ms2_df.groupby(groupby_columns).first().reset_index()
                result_df = result_df[kept_columns]

                ms2sum_df = ms2_df.groupby(groupby_columns).sum().reset_index()
                ms2norm_df = ms2_df.groupby(groupby_columns).max().reset_index()

                result_df["i"] = ms2sum_df["i"]
                result_df["i_norm"] = ms2norm_df["i_norm"]
                result_df["mslevel"] = 2

                # Calculating the MS1 i_norm and then joining on the ms1scan
                try:
                    ms1norm_df = ms1_df.groupby(groupby_columns).max().reset_index()
                    ms1norm_df["ms1scan"] = ms1norm_df["scan"]
                    ms1norm_df["i_norm_ms1"] = ms1norm_df["i_norm"]
                    ms1norm_df = ms1norm_df[["ms1scan", "i_norm_ms1"]]
                    result_df = result_df.merge(ms1norm_df, how="left", on="ms1scan")
                except:
                    pass

            return result_df

        if parsed_dict["querytype"]["function"] == "functionscanrangesum":
            result_list = []

            if parsed_dict["querytype"]["datatype"] == "datams1data":
                ms1_df["bin"] = ms1_df["mz"].apply(lambda x: int(x / 0.1))
                all_bins = set(ms1_df["bin"])

                for bin in all_bins:
                    ms1_filtered_df = ms1_df[ms1_df["bin"] == bin]
                    ms1sum_df = ms1_filtered_df.groupby("scan").sum().reset_index()

                    ms1_filtered_df = (
                        ms1_filtered_df.groupby("scan").first().reset_index()
                    )
                    ms1_filtered_df["i"] = ms1sum_df["i"]

                    result_list.append(ms1_filtered_df)

                return pd.concat(result_list)
            if parsed_dict["querytype"]["datatype"] == "datams2data":
                ms2_df = ms2_df.groupby("scan").sum()

                ms2sum_df = ms2_df.groupby("scan").sum()
                ms2_df = ms2_df.groupby("scan").first().reset_index()
                ms2_df["i"] = ms2sum_df["i"]

                return ms2_df

        print("APPLYING FUNCTION")    



import pandas as pd
from py_expression_eval import Parser
math_parser = Parser()

def _get_mz_tolerance(qualifiers, mz):
    if qualifiers is None:
        return 0.1

    if "qualifierppmtolerance" in qualifiers:
        ppm = qualifiers["qualifierppmtolerance"]["value"]
        mz_tol = abs(ppm * mz / 1000000)
        return mz_tol

    if "qualifiermztolerance" in qualifiers:
        return qualifiers["qualifiermztolerance"]["value"]

    return 0.1

def _get_massdefect_min(qualifiers):
    if qualifiers is None:
        return 0, 1

    if "qualifiermassdefect" in qualifiers:
        return qualifiers["qualifiermassdefect"]["min"], qualifiers["qualifiermassdefect"]["max"]
    
    return 0, 1


def _get_minintensity(qualifier):
    """
    Returns absolute min and relative min

    Args:
        qualifier ([type]): [description]

    Returns:
        [type]: [description]
    """

    min_intensity = 0
    min_percent_intensity = 0
    min_tic_percent_intensity = 0
    

    if qualifier is None:
        min_intensity = 0
        min_percent_intensity = 0

        return min_intensity, min_percent_intensity, min_tic_percent_intensity
    
    if "qualifierintensityvalue" in qualifier:
        min_intensity = float(qualifier["qualifierintensityvalue"]["value"])

    if "qualifierintensitypercent" in qualifier:
        min_percent_intensity = float(qualifier["qualifierintensitypercent"]["value"]) / 100

    if "qualifierintensityticpercent" in qualifier:
        min_tic_percent_intensity = float(qualifier["qualifierintensityticpercent"]["value"]) / 100

    # since the subsequent comparison is a strict greater than, if people set it to 100, then they won't get anything. 
    min_percent_intensity = min(min_percent_intensity, 0.99)

    return min_intensity, min_percent_intensity, min_tic_percent_intensity

def _get_exclusion_flag(qualifiers):
    if qualifiers is None:
        return False

    if "qualifierexcluded" in qualifiers:
        return True

    return False

def _set_intensity_register(ms_filtered_df, register_dict, condition):
    if "qualifiers" in condition:
        if "qualifierintensityreference" in condition["qualifiers"]:
            qualifier_variable = condition["qualifiers"]["qualifierintensitymatch"]["value"]

            grouped_df = ms_filtered_df.groupby("scan").sum().reset_index()
            for grouped_scan in grouped_df.to_dict(orient="records"):
                # Saving into the register
                key = "scan:{}:variable:{}".format(grouped_scan["scan"], qualifier_variable)
                register_dict[key] = grouped_scan["i"]
    return

def _filter_intensitymatch(ms_filtered_df, register_dict, condition):
    if "qualifiers" in condition:
        if "qualifierintensitymatch" in condition["qualifiers"] and \
            "qualifierintensitytolpercent" in condition["qualifiers"]:
            qualifier_expression = condition["qualifiers"]["qualifierintensitymatch"]["value"]
            qualifier_variable = qualifier_expression[0] #TODO: This assumes the variable is the first character in the expression, likely a bad assumption

            grouped_df = ms_filtered_df.groupby("scan").sum().reset_index()

            filtered_grouped_scans = []
            for grouped_scan in grouped_df.to_dict(orient="records"):
                # Reading from the register
                key = "scan:{}:variable:{}".format(grouped_scan["scan"], qualifier_variable)

                if key in register_dict:
                    register_value = register_dict[key]                    
                    evaluated_new_expression = math_parser.parse(qualifier_expression).evaluate({
                        qualifier_variable : register_value
                    })

                    min_match_intensity, max_match_intensity = _get_intensitymatch_range(condition["qualifiers"], evaluated_new_expression)

                    scan_intensity = grouped_scan["i"]

                    #print(key, scan_intensity, qualifier_expression, min_match_intensity, max_match_intensity, grouped_scan)

                    if scan_intensity > min_match_intensity and \
                        scan_intensity < max_match_intensity:
                        filtered_grouped_scans.append(grouped_scan)
                else:
                    # Its not in the register, which means we don't find it
                    continue
            return pd.DataFrame(filtered_grouped_scans)

    return ms_filtered_df

def _get_intensitymatch_range(qualifiers, match_intensity):
    """
    Matching the intensity range

    Args:
        qualifiers ([type]): [description]
        match_intensity ([type]): [description]

    Returns:
        [type]: [description]
    """

    min_intensity = 0
    max_intensity = 0

    if "qualifierintensitytolpercent" in qualifiers:
        tolerance_percent = qualifiers["qualifierintensitytolpercent"]["value"]
        tolerance_value = float(tolerance_percent) / 100 * match_intensity
        
        min_intensity = match_intensity - tolerance_value
        max_intensity = match_intensity + tolerance_value

    return min_intensity, max_intensity

def ms2prod_condition(condition, ms1_df, ms2_df, reference_conditions_register):
    """
    Filters the MS1 and MS2 data based upon MS2 peak conditions

    Args:
        condition ([type]): [description]
        ms1_df ([type]): [description]
        ms2_df ([type]): [description]
        reference_conditions_register ([type]): Edits this in place

    Returns:
        ms1_df ([type]): [description]
        ms2_df ([type]): [description]
    """
    exclusion_flag = _get_exclusion_flag(condition.get("qualifiers", None))

    if len(ms2_df) == 0:
        return ms1_df, ms2_df

    ms2_list = []
    for mz in condition["value"]:
        if mz == "ANY":
            # Checking defect options
            massdefect_min, massdefect_max = _get_massdefect_min(condition.get("qualifiers", None))
            ms2_filtered_df = ms2_df
            ms2_filtered_df["mz_defect"] = ms2_filtered_df["mz"] - ms2_filtered_df["mz"].astype(int)

            min_int, min_intpercent, min_tic_percent_intensity = _get_minintensity(condition.get("qualifiers", None))

            ms2_filtered_df = ms2_filtered_df[
                (ms2_filtered_df["mz_defect"] > massdefect_min) & 
                (ms2_filtered_df["mz_defect"] < massdefect_max) &
                (ms2_filtered_df["i"] > min_int) & 
                (ms2_filtered_df["i_norm"] > min_intpercent) & 
                (ms2_filtered_df["i_tic_norm"] > min_tic_percent_intensity)
            ]
        else:
            mz_tol = _get_mz_tolerance(condition.get("qualifiers", None), mz)
            mz_min = mz - mz_tol
            mz_max = mz + mz_tol

            min_int, min_intpercent, min_tic_percent_intensity = _get_minintensity(condition.get("qualifiers", None))

            ms2_filtered_df = ms2_df[(ms2_df["mz"] > mz_min) & 
                                    (ms2_df["mz"] < mz_max) & 
                                    (ms2_df["i"] > min_int) & 
                                    (ms2_df["i_norm"] > min_intpercent) & 
                                    (ms2_df["i_tic_norm"] > min_tic_percent_intensity)]

        # Setting the intensity match register
        _set_intensity_register(ms2_filtered_df, reference_conditions_register, condition)

        # Applying the intensity match
        ms2_filtered_df = _filter_intensitymatch(ms2_filtered_df, reference_conditions_register, condition)

        ms2_list.append(ms2_filtered_df)

    if len(ms2_list) == 1:
        ms2_filtered_df = ms2_list[0]
    else:
        ms2_filtered_df = pd.concat(ms2_list)

    # Apply the negation operator
    if exclusion_flag:
        filtered_scans = set(ms2_filtered_df["scan"])
        original_scans = set(ms2_df["scan"])
        negation_scans = original_scans - filtered_scans

        ms2_filtered_df = ms2_df[ms2_df["scan"].isin(negation_scans)]

    if len(ms2_filtered_df) == 0:
       return pd.DataFrame(), pd.DataFrame()
    
    # Filtering the actual data structures
    filtered_scans = set(ms2_filtered_df["scan"])
    ms2_df = ms2_df[ms2_df["scan"].isin(filtered_scans)]

    # Filtering the MS1 data now
    ms1_scans = set(ms2_df["ms1scan"])
    ms1_df = ms1_df[ms1_df["scan"].isin(ms1_scans)]

    return ms1_df, ms2_df

def ms2nl_condition(condition, ms1_df, ms2_df, reference_conditions_register):
    """
    Filters the MS1 and MS2 data based upon MS2 neutral loss conditions

    Args:
        condition ([type]): [description]
        ms1_df ([type]): [description]
        ms2_df ([type]): [description]
        reference_conditions_register ([type]): Edits this in place

    Returns:
        ms1_df ([type]): [description]
        ms2_df ([type]): [description]
    """
    exclusion_flag = _get_exclusion_flag(condition.get("qualifiers", None))

    if len(ms2_df) == 0:
        return ms1_df, ms2_df

    ms2_list = []
    for mz in condition["value"]:
        if mz == "ANY":
            # Checking defect options
            massdefect_min, massdefect_max = _get_massdefect_min(condition.get("qualifiers", None))
            ms2_filtered_df = ms2_df
            ms2_filtered_df["mz_defect"] = ms2_filtered_df["mz"] - ms2_filtered_df["mz"].astype(int)

            min_int, min_intpercent, min_tic_percent_intensity = _get_minintensity(condition.get("qualifiers", None))

            ms2_filtered_df = ms2_filtered_df[
                (ms2_filtered_df["mz_defect"] > massdefect_min) & 
                (ms2_filtered_df["mz_defect"] < massdefect_max) &
                (ms2_filtered_df["i"] > min_int) & 
                (ms2_filtered_df["i_norm"] > min_intpercent) & 
                (ms2_filtered_df["i_tic_norm"] > min_tic_percent_intensity)
            ]
        else:
            mz_tol = _get_mz_tolerance(condition.get("qualifiers", None), mz) #TODO: This is incorrect logic if it comes to PPM accuracy
            nl_min = mz - mz_tol
            nl_max = mz + mz_tol

            min_int, min_intpercent, min_tic_percent_intensity = _get_minintensity(condition.get("qualifiers", None))

            ms2_filtered_df = ms2_df[
                ((ms2_df["precmz"] - ms2_df["mz"]) > nl_min) & 
                ((ms2_df["precmz"] - ms2_df["mz"]) < nl_max) &
                (ms2_df["i"] > min_int) & 
                (ms2_df["i_norm"] > min_intpercent) & 
                (ms2_df["i_tic_norm"] > min_tic_percent_intensity)
            ]

        # Setting the intensity match register
        _set_intensity_register(ms2_filtered_df, reference_conditions_register, condition)

        # Applying the intensity match
        ms2_filtered_df = _filter_intensitymatch(ms2_filtered_df, reference_conditions_register, condition)

        ms2_list.append(ms2_filtered_df)

    if len(ms2_list) == 1:
        ms2_filtered_df = ms2_list[0]
    else:
        ms2_filtered_df = pd.concat(ms2_list)

    # Apply the negation operator
    if exclusion_flag:
        filtered_scans = set(ms2_filtered_df["scan"])
        original_scans = set(ms2_df["scan"])
        negation_scans = original_scans - filtered_scans

        ms2_filtered_df = ms2_df[ms2_df["scan"].isin(negation_scans)]

    if len(ms2_filtered_df) == 0:
       return pd.DataFrame(), pd.DataFrame()

    # Filtering the actual data structures
    filtered_scans = set(ms2_filtered_df["scan"])
    ms2_df = ms2_df[ms2_df["scan"].isin(filtered_scans)]

    # Filtering the MS1 data now
    ms1_scans = set(ms2_df["ms1scan"])
    ms1_df = ms1_df[ms1_df["scan"].isin(ms1_scans)]

    return ms1_df, ms2_df

def ms2prec_condition(condition, ms1_df, ms2_df, reference_conditions_register):
    """
    Filters the MS1 and MS2 data based upon MS2 precursor conditions

    Args:
        condition ([type]): [description]
        ms1_df ([type]): [description]
        ms2_df ([type]): [description]
        reference_conditions_register ([type]): Edits this in place

    Returns:
        ms1_df ([type]): [description]
        ms2_df ([type]): [description]
    """
    exclusion_flag = _get_exclusion_flag(condition.get("qualifiers", None))

    if len(ms2_df) == 0:
        return ms1_df, ms2_df

    ms2_list = []
    for mz in condition["value"]:
        if mz == "ANY":
            # Checking defect options
            massdefect_min, massdefect_max = _get_massdefect_min(condition.get("qualifiers", None))
            ms2_filtered_df = ms2_df
            ms2_filtered_df["precmz_defect"] = ms2_filtered_df["precmz"] - ms2_filtered_df["precmz"].astype(int)

            ms2_filtered_df = ms2_filtered_df[(
                ms2_filtered_df["precmz_defect"] > massdefect_min) & 
                (ms2_filtered_df["precmz_defect"] < massdefect_max)
            ]
        else:
            mz_tol = _get_mz_tolerance(condition.get("qualifiers", None), mz)
            mz_min = mz - mz_tol
            mz_max = mz + mz_tol

            ms2_filtered_df = ms2_df[(
                ms2_df["precmz"] > mz_min) & 
                (ms2_df["precmz"] < mz_max)
            ]

        ms2_list.append(ms2_filtered_df)

    if len(ms2_list) == 1:
        ms2_filtered_df = ms2_list[0]
    else:
        ms2_filtered_df = pd.concat(ms2_list)
    
    # Apply the negation operator
    if exclusion_flag:
        filtered_scans = set(ms2_filtered_df["scan"])
        original_scans = set(ms2_df["scan"])
        negation_scans = original_scans - filtered_scans

        ms2_filtered_df = ms2_df[ms2_df["scan"].isin(negation_scans)]

    if len(ms2_filtered_df) == 0:
       return pd.DataFrame(), pd.DataFrame()
    
    # Filtering the actual data structures
    filtered_scans = set(ms2_filtered_df["scan"])
    ms2_df = ms2_df[ms2_df["scan"].isin(filtered_scans)]

    # Filtering the MS1 data now
    if len(ms1_df) > 0:
        ms1_scans = set(ms2_df["ms1scan"])
        ms1_df = ms1_df[ms1_df["scan"].isin(ms1_scans)]

    return ms1_df, ms2_df

def ms1_condition(condition, ms1_df, ms2_df, reference_conditions_register):
    """
    Filters the MS1 and MS2 data based upon MS1 peak conditions

    Args:
        condition ([type]): [description]
        ms1_df ([type]): [description]
        ms2_df ([type]): [description]
        reference_conditions_register ([type]): Edits this in place

    Returns:
        ms1_df ([type]): [description]
        ms2_df ([type]): [description]
    """
    exclusion_flag = _get_exclusion_flag(condition.get("qualifiers", None))

    if len(ms1_df) == 0:
        return ms1_df, ms2_df

    ms1_list = []
    for mz in condition["value"]:
        if mz == "ANY":
            # Checking defect options
            massdefect_min, massdefect_max = _get_massdefect_min(condition.get("qualifiers", None))
            ms1_filtered_df = ms1_df
            ms1_filtered_df["mz_defect"] = ms1_filtered_df["mz"] - ms1_filtered_df["mz"].astype(int)

            min_int, min_intpercent, min_tic_percent_intensity = _get_minintensity(condition.get("qualifiers", None))

            ms1_filtered_df = ms1_filtered_df[
                (ms1_filtered_df["mz_defect"] > massdefect_min) & 
                (ms1_filtered_df["mz_defect"] < massdefect_max) &
                (ms1_filtered_df["i"] > min_int) & 
                (ms1_filtered_df["i_norm"] > min_intpercent) & 
                (ms1_filtered_df["i_tic_norm"] > min_tic_percent_intensity)
            ]
        else:
            mz_tol = _get_mz_tolerance(condition.get("qualifiers", None), mz)
            mz_min = mz - mz_tol
            mz_max = mz + mz_tol

            min_int, min_intpercent, min_tic_percent_intensity = _get_minintensity(condition.get("qualifiers", None))
            ms1_filtered_df = ms1_df[
                (ms1_df["mz"] > mz_min) & 
                (ms1_df["mz"] < mz_max) & 
                (ms1_df["i"] > min_int) & 
                (ms1_df["i_norm"] > min_intpercent) & 
                (ms1_df["i_tic_norm"] > min_tic_percent_intensity)]

        # Setting the intensity match register
        _set_intensity_register(ms1_filtered_df, reference_conditions_register, condition)

        # Applying the intensity match
        ms1_filtered_df = _filter_intensitymatch(ms1_filtered_df, reference_conditions_register, condition)

        ms1_list.append(ms1_filtered_df)
    
    if len(ms1_list) == 1:
        ms1_filtered_df = ms1_list[0]
    else:
        ms1_filtered_df = pd.concat(ms1_list)

    # Apply the negation operator
    if exclusion_flag:
        filtered_scans = set(ms1_filtered_df["scan"])
        original_scans = set(ms1_df["scan"])
        negation_scans = original_scans - filtered_scans

        ms1_filtered_df = ms1_df[ms1_df["scan"].isin(negation_scans)]

    if len(ms1_filtered_df) == 0:
       return pd.DataFrame(), pd.DataFrame()

    # Filtering the actual data structures
    filtered_scans = set(ms1_filtered_df["scan"])
    ms1_df = ms1_df[ms1_df["scan"].isin(filtered_scans)]

    if "ms1scan" in ms2_df:
        ms2_df = ms2_df[ms2_df["ms1scan"].isin(filtered_scans)]

    return ms1_df, ms2_df

    



from massql import msql_parser
from massql import msql_engine
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from py_expression_eval import Parser
math_parser = Parser()


def visualize_query(query, variable_x=500, variable_y=1, precursor_mz=800, ms1_peaks=None, ms2_peaks=None):
    """
    This function creates a plotly visualization. If ms1 and ms2 peaks are included, we will plot them along with it. 

    Args:
        query ([type]): [description]
        variable_x (int, optional): [description]. Defaults to 500.
        variable_y (int, optional): [description]. Defaults to 1.
        precursor_mz (int, optional): [description]. Defaults to 800.
        ms1_peaks ([type], optional): [description]. Defaults to None.
        ms2_peaks ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """

    parsed_query = msql_parser.parse_msql(query)

    # Cleaning up variables
    for condition in parsed_query["conditions"]:
        # Setting m/z variables
        if "value" in condition:
            for i, value in enumerate(condition["value"]):
                try:
                    # Checking if X is in any string
                    if "X" in value:
                        condition["value"][i] = math_parser.parse(value).evaluate({
                                    "X" : variable_x
                                })    
                except:
                    pass

        # Setting intensity variables
        if "qualifiers" in condition:
            if "qualifierintensitymatch" in condition["qualifiers"]:
                value = condition["qualifiers"]["qualifierintensitymatch"]["value"]
                condition["qualifiers"]["qualifierintensitymatch"]["value"] = math_parser.parse(value).evaluate({
                                "Y" : variable_y,
                                "X" : variable_x
                            })

    ms1_fig = go.Figure()
    ms2_fig = go.Figure()

    ms1_min_x_axis = 0
    ms1_max_x_axis = 1000

    ms2_min_x_axis = 0
    ms2_max_x_axis = 1000

    if ms1_peaks is not None:
        max_int = max([peak[1] for peak in ms1_peaks])
        # Drawing the spectrum object
        mzs = [peak[0] for peak in ms1_peaks]
        ints = [peak[1]/max_int for peak in ms1_peaks]
        neg_ints = [intensity * -1 for intensity in ints]

        ms1_max_x_axis = max(ms1_max_x_axis, max(mzs))

        # Hover data
        hover_labels = ["{:.4f} m/z, {:.2f} int".format(mzs[i], ints[i]) for i in range(len(mzs))]

        ms1_fig = go.Figure(
            data=go.Scatter(x=mzs, y=ints, 
                mode='markers',
                marker=dict(size=0.00001),
                error_y=dict(
                    symmetric=False,
                    arrayminus=[0]*len(neg_ints),
                    array=neg_ints,
                    width=0
                ),
                text=hover_labels,
                textposition="top right"
            )
        )


    if ms2_peaks is not None:
        max_int = max([peak[1] for peak in ms2_peaks])
        # Drawing the spectrum object
        mzs = [peak[0] for peak in ms2_peaks]
        ints = [peak[1]/max_int for peak in ms2_peaks]
        neg_ints = [intensity * -1 for intensity in ints]

        ms2_max_x_axis = max(ms2_max_x_axis, max(mzs))

        # Hover data
        hover_labels = ["{:.4f} m/z, {:.3f} int".format(mzs[i], ints[i]) for i in range(len(mzs))]

        ms2_fig = go.Figure(
            data=go.Scatter(x=mzs, y=ints, 
                mode='markers',
                marker=dict(size=0.00001),
                error_y=dict(
                    symmetric=False,
                    arrayminus=[0]*len(neg_ints),
                    array=neg_ints,
                    width=0
                ),
                text=hover_labels,
                textposition="top right"
            )
        )

    for condition in parsed_query["conditions"]:
        if condition["conditiontype"] != "where":
            continue

        if condition["type"] == "ms2productcondition" and condition["conditiontype"] == "where":
            mz = condition["value"][0]
            mz_tol = msql_engine._get_mz_tolerance(condition.get("qualifiers", None), mz)

            mz_min = mz - mz_tol
            mz_max = mz + mz_tol

            intensity = 1

            if "qualifiers" in condition:
                if "qualifierintensitymatch" in condition["qualifiers"]:
                    intensity = condition["qualifiers"]["qualifierintensitymatch"]["value"]

            # Draw the bounds of the for the MS2 Product
            ms2_fig.add_shape(type="rect",
                x0=mz_min, y0=0, x1=mz_max, y1=intensity,
                line=dict(
                    color="Green",
                    width=2,
                )
            )

        if condition["type"] == "ms2neutrallosscondition" and condition["conditiontype"] == "where":
            mz = condition["value"][0]
            mz_tol = msql_engine._get_mz_tolerance(condition.get("qualifiers", None), mz)
            nl_min = mz - mz_tol
            nl_max = mz + mz_tol

            mz_min = precursor_mz - nl_max
            mz_max = precursor_mz - nl_min

            intensity = 1

            if "qualifiers" in condition:
                if "qualifierintensitymatch" in condition["qualifiers"]:
                    intensity = condition["qualifiers"]["qualifierintensitymatch"]["value"]

            ms2_fig.add_shape(type="rect",
                x0=mz_min, y0=0, x1=mz_max, y1=intensity,
                line=dict(
                    color="Red",
                    width=2,
                )
            )
        
        if condition["type"] == "xcondition":
            print(condition)
            ms1_fig.add_shape(type="line",
                x0=condition["min"], y0=0, x1=condition["min"], y1=1,
                line=dict(
                    color="pink",
                    width=2,
                    dash="dot",
                )
            )

            ms1_fig.add_shape(type="line",
                x0=condition["max"], y0=0, x1=condition["max"], y1=1,
                line=dict(
                    color="pink",
                    width=2,
                    dash="dot",
                )
            )

        if condition["type"] == "ms1mzcondition":
            mz = condition["value"][0]
            mz_tol = msql_engine._get_mz_tolerance(condition.get("qualifiers", None), mz)
            mz_min = mz - mz_tol
            mz_max = mz + mz_tol

            intensity = 1

            if "qualifiers" in condition:
                if "qualifierintensitymatch" in condition["qualifiers"]:
                    intensity = condition["qualifiers"]["qualifierintensitymatch"]["value"]

            ms1_fig.add_shape(type="rect",
                x0=mz_min, y0=0, x1=mz_max, y1=intensity,
                line=dict(
                    color="Green",
                    width=2,
                )
            )

            # Determining if we should draw intensity bounds
            if "qualifiers" in condition:
                if "qualifierintensitytolpercent" in condition["qualifiers"]:
                    percent_tolerance = condition["qualifiers"]["qualifierintensitytolpercent"]["value"]
                    percent_tolerance = percent_tolerance / 100
                    min_intensity = intensity - intensity * percent_tolerance
                    max_intensity = intensity + intensity * percent_tolerance
                    intensity_gap = max_intensity - min_intensity
                    mz_gap = mz_max - mz_min

                    # Top and bottom bounds

                    ms1_fig.add_shape(type="line",
                        x0=mz_min, y0=min_intensity, x1=mz_max, y1=min_intensity,
                        line=dict(
                            color="purple",
                            width=2,
                            dash="dot",
                        )
                    )

                    ms1_fig.add_shape(type="line",
                        x0=mz_min, y0=max_intensity, x1=mz_max, y1=max_intensity,
                        line=dict(
                            color="purple",
                            width=2,
                            dash="dot",
                        )
                    )

                    # Drawing arrows
                    
                    # Down Arrow
                    ms1_fig.add_annotation(
                        x=mz_min + mz_gap / 5,  # arrows' head
                        y=max_intensity - intensity_gap / 4,  # arrows' head
                        ax=mz_min + mz_gap / 5,  # arrows' tail
                        ay=max_intensity,  # arrows' tail
                        xref='x',
                        yref='y',
                        axref='x',
                        ayref='y',
                        text='',  # if you want only the arrow
                        showarrow=True,
                        arrowhead=3,
                        arrowsize=1,
                        arrowwidth=1,
                        arrowcolor='purple'
                    )

                    # Up Arrow
                    ms1_fig.add_annotation(
                        x=mz_min + mz_gap / 5,  # arrows' head
                        y=min_intensity + intensity_gap / 4,  # arrows' head
                        ax=mz_min + mz_gap / 5,  # arrows' tail
                        ay=min_intensity,  # arrows' tail
                        xref='x',
                        yref='y',
                        axref='x',
                        ayref='y',
                        text='',  # if you want only the arrow
                        showarrow=True,
                        arrowhead=3,
                        arrowsize=1,
                        arrowwidth=1,
                        arrowcolor='purple'
                    )

                if "qualifierintensitypercent" in condition["qualifiers"]:
                    value = condition["qualifiers"]["qualifierintensitypercent"]["value"] / 100
                    min_intensity = value
                    mz_gap = mz_max - mz_min

                    ms1_fig.add_shape(type="line",
                        x0=mz_min, y0=value, x1=mz_max, y1=value,
                        line=dict(
                            color="red",
                            width=2,
                            dash="dot",
                        )
                    )

                    # Up Arrow
                    ms1_fig.add_annotation(
                        x=mz_min + mz_gap / 5,  # arrows' head
                        y=min_intensity + 0.2 * min_intensity,  # arrows' head
                        ax=mz_min + mz_gap / 5,  # arrows' tail
                        ay=min_intensity,  # arrows' tail
                        xref='x',
                        yref='y',
                        axref='x',
                        ayref='y',
                        text='',  # if you want only the arrow
                        showarrow=True,
                        arrowhead=3,
                        arrowsize=1,
                        arrowwidth=1,
                        arrowcolor='red'
                    )


    # Set axes properties
    ms2_fig.update_xaxes(range=[ms2_min_x_axis, ms2_max_x_axis], showgrid=False)
    ms2_fig.update_yaxes(range=[0, 1.5])

    ms2_fig.update_layout(
        title="MS2 Query Visualization, Precursor m/z {}".format(precursor_mz),
        xaxis_title="m/z",
        yaxis_title="intensity",
    )

    # Setting axes properties
    ms1_fig.update_xaxes(range=[ms1_min_x_axis, ms1_max_x_axis], showgrid=False)
    ms1_fig.update_yaxes(range=[0, 1.5])

    ms1_fig.update_layout(
        title="MS1 Query Visualization",
        xaxis_title="m/z",
        yaxis_title="intensity",
    )

    return ms1_fig, ms2_fig


from massql.msql_engine import _executeconditions_query, _executecollate_query
import ray

@ray.remote
def _executeconditions_query_ray(parsed_dict_list, input_filename, ms1_input_df=None, ms2_input_df=None, cache=True):
    """
    Here we will use parallel ray, we will give a list of dictionaries to query, and return a list of results that are collated

    Args:
        parsed_dict_list ([type]): [description]
        input_filename ([type]): [description]
        ms1_input_df ([type], optional): [description]. Defaults to None.
        ms2_input_df ([type], optional): [description]. Defaults to None.
        cache (bool, optional): [description]. Defaults to True.

    Returns:
        [type]: [description]
    """

    collated_list = []

    for parsed_dict in parsed_dict_list:
        ms1_df, ms2_df = _executeconditions_query(parsed_dict, input_filename, ms1_input_df=ms1_input_df, ms2_input_df=ms2_input_df, cache=cache)

        collated_df = _executecollate_query(parsed_dict, ms1_df, ms2_df)
        collated_list.append(collated_df)

    return collated_list


import json
import os
import argparse
import glob
import sys
import pandas as pd
import numpy as np
import pymzml
from pyteomics import mzxml
from matchms.importing import load_from_mgf
from psims.mzml.writer import MzMLWriter

def main():
    parser = argparse.ArgumentParser(description="MSQL Query in Proteosafe")
    parser.add_argument('input_folder', help='Input filename')
    parser.add_argument('results_file', help='Input Query')
    parser.add_argument('extracted_mgf', help='extracted_mgf')
    parser.add_argument('extracted_mzML', help='extracted_mgf')
    parser.add_argument('extracted_result', help='extracted_mgf')

    args = parser.parse_args()

    if os.path.isdir(args.results_file):
        results_df = pd.read_csv(glob.glob(os.path.join(args.results_file, "*"))[0], sep='\t')
    else:
        results_df = pd.read_csv(args.results_file, sep='\t')

    try:
        _extract_spectra(results_df, 
                        args.input_folder, 
                        output_mgf_filename=args.extracted_mgf,
                        output_mzML_filename=args.extracted_mzML,
                        output_summary=args.extracted_result)
    except Exception as e: 
        print("FAILURE ON EXTRACTION")
        pass

def _extract_mzML_scan(input_filename, spectrum_identifier_list):
    MS_precisions = {
        1: 5e-6,
        2: 20e-6,
        3: 20e-6,
        4: 20e-6,
        5: 20e-6,
        6: 20e-6,
        7: 20e-6,
    }
    
    output_list = []

    spectrum_identifier_set = set([str(spectrum_scan) for spectrum_scan in spectrum_identifier_list])

    run = pymzml.run.Reader(input_filename, MS_precisions=MS_precisions)

    
    for spec in run:
        if str(spec.ID) in spectrum_identifier_set:
            try:
                peaks = spec.peaks("raw")

                # Filtering out zero rows
                peaks = peaks[~np.any(peaks < 1.0, axis=1)]

                # Sorting by intensity
                peaks = peaks[peaks[:, 1].argsort()]

                # Getting top 1000
                #peaks = peaks[-1000:]

                if len(peaks) == 0:
                    return None

                mz, intensity = zip(*peaks)

                mz_list = list(mz)
                i_list = list(intensity)

                peaks_list = []
                for i in range(len(mz_list)):
                    peaks_list.append([float(mz_list[i]), float(i_list[i])])

                # Sorting Peaks
                peaks_list = sorted(peaks_list, key=lambda x: x[0])

                spectrum_obj = {}
                spectrum_obj["peaks"] = peaks_list
                spectrum_obj["mslevel"] = spec.ms_level
                spectrum_obj["scan"] = str(spec.ID)

                if spec.ms_level > 1:
                    msn_mz = spec.selected_precursors[0]["mz"]
                    spectrum_obj["precursor_mz"] = msn_mz

                output_list.append(spectrum_obj)
            except:
                raise

    return output_list

def _extract_mzXML_scan(input_filename, spectrum_identifier_list):
    output_list = []
    spectrum_identifier_set = set([str(spectrum_scan) for spectrum_scan in spectrum_identifier_list])

    with mzxml.read(input_filename) as reader:
        for spectrum in reader:
            if str(spectrum["id"]) in spectrum_identifier_set:
                spec = spectrum

                mz_list = list(spec["m/z array"])
                i_list = list(spec["intensity array"])

                peaks_list = []
                for i in range(len(mz_list)):
                    peaks_list.append([float(mz_list[i]), float(i_list[i])])

                # Sorting Peaks
                peaks_list = sorted(peaks_list, key=lambda x: x[0])

                # Loading Data
                spectrum_obj = {}
                spectrum_obj["peaks"] = peaks_list
                spectrum_obj["mslevel"] = spec["msLevel"]
                spectrum_obj["scan"] = str(spectrum["id"])

                if spec["msLevel"] > 1:
                    msn_mz = spec["precursorMz"][0]["precursorMz"]
                    spectrum_obj["precursor_mz"] = msn_mz

                output_list.append(spectrum_obj)

    return output_list

def _extract_mgf_scan(input_filename, spectrum_identifier_list):
    output_list = []
    spectrum_identifier_set = set([str(spectrum_scan) for spectrum_scan in spectrum_identifier_list])

    file = load_from_mgf(input_filename)

    spec = None
    for spectrum in file:
        scan_number = spectrum.metadata["scans"]
        if str(scan_number) in spectrum_identifier_set:
            spec = spectrum

            mz_list = list(spec.peaks.mz)
            i_list = list(spec.peaks.intensities)

            peaks_list = []
            for i in range(len(mz_list)):
                peaks_list.append([mz_list[i], i_list[i]])

            # Loading Data
            spectrum_obj = {}
            spectrum_obj["peaks"] = peaks_list
            spectrum_obj["mslevel"] = 2
            spectrum_obj["scan"] = scan_number
            spectrum_obj["precursor_mz"] = float(spec.metadata["pepmass"][0])

            output_list.append(spectrum_obj)

    return output_list


def _extract_spectra(results_df, input_spectra_folder, 
                    output_mgf_filename=None, 
                    output_mzML_filename=None, 
                    output_json_filename=None,
                    output_summary=None):
    spectrum_list = []
    result_df_list = []

    # TODO: reduce duplicate scans to extract

    # Lets group by file name so we only have to load the files once
    grouped_results_df = results_df.groupby("filename")

    current_scan = 1
    for filename, results_by_file_df in grouped_results_df:

        try:
            if "mangled_filename" in results_by_file_df:
                input_spectra_filename = os.path.join(input_spectra_folder, results_by_file_df["mangled_filename"].iloc[0])
            else:
                input_spectra_filename = os.path.join(input_spectra_folder, results_by_file_df["filename"].iloc[0])

            spectrum_obj_list = []
            if ".mzML" in input_spectra_filename:
                spectrum_obj_list = _extract_mzML_scan(input_spectra_filename, list(set(results_by_file_df["scan"])))
            if ".mzXML" in input_spectra_filename:
                spectrum_obj_list = _extract_mzXML_scan(input_spectra_filename, list(set(results_by_file_df["scan"])))
            if ".mgf" in input_spectra_filename:
                spectrum_obj_list = _extract_mgf_scan(input_spectra_filename, list(set(results_by_file_df["scan"])))


            for spectrum_obj in spectrum_obj_list:                
                # These are a new scan number in the file, not sure if we need this
                spectrum_obj["new_scan"] = current_scan

                filtered_by_scan_df = results_by_file_df[results_by_file_df["scan"].astype(str) == str(spectrum_obj["scan"])]
                filtered_by_scan_df["new_scan"] = current_scan
                result_df_list.append(filtered_by_scan_df)

                spectrum_obj["query_results"] = filtered_by_scan_df.to_dict(orient="records")
                spectrum_list.append(spectrum_obj)
                current_scan += 1

        except KeyboardInterrupt:
            raise
        except:
            print("Error", filename)
            pass

    merged_summary_df = pd.concat(result_df_list)
    # Writing the updated extraction
    if output_summary is not None:
        merged_summary_df.to_csv(output_summary, sep='\t', index=False)

    if len(spectrum_list) > 5000:
        print("Not Extracting, too many spectra")
        return None

    if output_json_filename is not None:
        with open(output_json_filename, "w") as o:
            o.write(json.dumps(spectrum_list))

    # Writing the spectrum now
    if output_mgf_filename is not None:
        _export_mgf(spectrum_list, output_mgf_filename)

    if output_mzML_filename is not None:
        _export_mzML(spectrum_list, output_mzML_filename)

    return merged_summary_df

    

def _export_mgf(spectrum_list, output_mgf_filename):
    with open(output_mgf_filename, "w") as o:
        for i, spectrum in enumerate(spectrum_list):
            o.write("BEGIN IONS\n")
            if "precursor_mz" in spectrum:
                o.write("PEPMASS={}\n".format(spectrum["precursor_mz"]))
            o.write("SCANS={}\n".format(spectrum["new_scan"]))
            for peak in spectrum["peaks"]:
                o.write("{} {}\n".format(peak[0], peak[1]))
            o.write("END IONS\n")

def _export_mzML(spectrum_list, output_mzML_filename):
    with MzMLWriter(open(output_mzML_filename, 'wb'), close=True) as out:
        # Add default controlled vocabularies
        out.controlled_vocabularies()
        # Open the run and spectrum list sections
        with out.run(id="my_analysis"):
            spectrum_count = len(spectrum_list)
            with out.spectrum_list(count=spectrum_count):
                for spectrum in spectrum_list:
                    mz_list = [peak[0] for peak in spectrum["peaks"]]
                    i_list = [peak[1] for peak in spectrum["peaks"]]
                    mslevel = spectrum["mslevel"]

                    if mslevel == 1:
                        # Write Precursor scan
                        out.write_spectrum(
                            mz_list, i_list,
                            id="scan={}".format(spectrum["new_scan"]), params=[
                                "MS1 Spectrum",
                                {"ms level": 1},
                                {"total ion current": sum(i_list)}
                            ])
                    elif mslevel == 2:
                        out.write_spectrum(
                            mz_list, i_list,
                            id="scan={}".format(spectrum["new_scan"]), params=[
                                "MSn Spectrum",
                                {"ms level": 2},
                                {"total ion current": sum(i_list)}
                            ],
                            # Include precursor information
                            precursor_information={
                                "mz": spectrum["precursor_mz"],
                                "intensity": 0,
                                "charge": 0,
                                "scan_id": 0,
                                "activation": ["beam-type collisional dissociation", {"collision energy": 25}],
                                "isolation_window": [spectrum["precursor_mz"] - 1, spectrum["precursor_mz"], spectrum["precursor_mz"] + 1]
                            })


if __name__ == "__main__":
    main()



import os

from lark import Lark
from lark import Transformer
from lark import tree

from py_expression_eval import Parser
math_parser = Parser()

from pyteomics import mass



#TODO: Update language definition to make it such that we can distinguish different functions

class MassQLToJSON(Transformer):
   def wherekeyword(self, items):
      return "WHERE"

   def querykeyword(self, items):
      return "QUERY"

   def booleanandconjunction(self, s):
      return "AND"

   def booleanorconjunction(self, s):
      return "OR"

   def qualifiermztolerance(self, items):
      return "qualifiermztolerance"
   
   def qualifierppmtolerance(self, items):
      return "qualifierppmtolerance"

   def qualifierintensityvalue(self, items):
      return "qualifierintensityvalue"

   def qualifierintensitypercent(self, items):
      return "qualifierintensitypercent"

   def qualifierintensityticpercent(self, items):
      return "qualifierintensityticpercent"

   def qualifierintensitymatch(self, items):
      return "qualifierintensitymatch"

   def qualifierintensitytolpercent(self, items):
      return "qualifierintensitytolpercent"

   def qualifierintensityreference(self, items):
      return "qualifierintensityreference"
   
   def qualifierexclude(self, items):
      return "qualifierexcluded"

   def ms2productcondition(self, items):
      return "ms2productcondition"

   def ms2precursorcondition(self, items):
      return "ms2precursorcondition"

   def ms2neutrallosscondition(self, items):
      return "ms2neutrallosscondition"
   
   def ms1mzcondition(self, items):
      return "ms1mzcondition"
   
   def rtmincondition(self, items):
      return "rtmincondition"

   def rtmaxcondition(self, items):
      return "rtmaxcondition"

   def scanmincondition(self, items):
      return "scanmincondition"

   def scanmaxcondition(self, items):
      return "scanmaxcondition"

   def chargecondition(self, items):
      return "chargecondition"

   def polaritycondition(self, items):
      return "polaritycondition"

   def positivepolarity(self, items):
      return "positivepolarity"

   def negativepolarity(self, items):
      return "negativepolarity"

   def xcondition(self, items):
      return "xcondition"

   def xfunction(self, items):
      return items[0]
   
   def xrange(self, items):
      return "xrange"
   
   def xdefect(self, items):
      return "xdefect"

   def qualifiermassdefect(self, items):
      return "qualifiermassdefect"

   # Handling mobility literals
   def mobilitycondition(self, items):
      return "mobilitycondition"

   def mobilityfunction(self, items):
      return items[0]
   
   def mobilityrange(self, items):
      return "mobilityrange"

   def qualifier(self, items):
      if len(items) == 1 and items[0] == "qualifierintensityreference":
         qualifier_type = items[0]
         
         qualifier_dict = {}
         qualifier_dict["type"] = "qualifier"
         qualifier_dict[qualifier_type] = {}
         qualifier_dict[qualifier_type]["name"] = qualifier_type
      
      if len(items) == 1 and items[0] == "qualifierexcluded":
         qualifier_type = items[0]
         
         qualifier_dict = {}
         qualifier_dict["type"] = "qualifier"
         qualifier_dict[qualifier_type] = {}
         qualifier_dict[qualifier_type]["name"] = qualifier_type

      # We are at a qualifier leaf
      if len(items) == 3:
         qualifier_type = items[0]

         qualifier_dict = {}
         qualifier_dict["type"] = "qualifier"
         qualifier_dict[qualifier_type] = {}
         qualifier_dict[qualifier_type]["name"] = qualifier_type

         if qualifier_type == "qualifierppmtolerance":
            qualifier_dict[qualifier_type]["unit"] = "ppm"
         if qualifier_type == "qualifiermztolerance":
            qualifier_dict[qualifier_type]["unit"] = "mz"

         if qualifier_dict[qualifier_type]["name"] == "qualifierintensitymatch":
            # NOTE: these matches should contain variables that the engine will understand
            qualifier_dict[qualifier_type]["value"] = items[-1]
         else:
            qualifier_dict[qualifier_type]["value"] = float(items[-1])

      # We are at a merge node for the qualifier
      if len(items) == 2:
         qualifier_dict = {}
         
         for qualifier in items:
            for key in qualifier:
               qualifier_dict[key] = qualifier[key]

      if len(items) == 5:
         # This is mass defect
         qualifier_type = items[0]

         if qualifier_type == "qualifiermassdefect":
            qualifier_dict = {}
            qualifier_dict["type"] = "qualifier"
            qualifier_dict[qualifier_type] = {}
            qualifier_dict[qualifier_type]["name"] = qualifier_type
            qualifier_dict[qualifier_type]["min"] = items[-2]
            qualifier_dict[qualifier_type]["max"] = items[-1]

      return qualifier_dict

   def conditionfields(self, items):
      return items[0]

   def condition(self, items):
      condition_type = items[0]

      if len(items) == 3:
         # These are for or numerical expression polarity or numerical expression
         if condition_type == "polaritycondition":
            condition_dict = {}
            condition_dict["type"] = items[0]
            condition_dict["value"] = [items[-1]]
         else:
            # These are numerical expressions for mz,rt type conditions
            # This might be an item or could be a list
            if isinstance(items[-1], list):
               condition_dict = {}
               condition_dict["type"] = items[0]
               condition_dict["value"] = items[-1]
            else:
               condition_dict = {}
               condition_dict["type"] = items[0]
               condition_dict["value"] = [items[-1]]
      elif len(items) == 5:
         # These are for the x range clauses
         if items[0] == "xcondition":
            function = items[2]

            condition_dict = {}
            condition_dict["type"] = condition_type

            if function == "xdefect":
               condition_dict["mindefect"] = float(items[-2])
               condition_dict["maxdefect"] = float(items[-1])
            elif function == "xrange":
               condition_dict["min"] = float(items[-2])
               condition_dict["max"] = float(items[-1])

         elif items[0] == "mobilitycondition":
            function = items[2]

            condition_dict = {}
            condition_dict["type"] = condition_type

            if function == "mobilityrange":
               condition_dict["min"] = items[-2]
               condition_dict["max"] = items[-1]
   
      return condition_dict

   def wherefullcondition(self, items):
      """
      Defines the full set of qualifiers for a single constraint or all constraints

      Args:
          items ([type]): [description]

      Returns:
          [type]: [description]
      """

      # Only condition, no qualifiers
      if len(items) == 1:
         items[0]["conditiontype"] = "where"
         return items
      
      # Has potentially a qualifier
      if len(items) == 2:
         if items[1]["type"] == "qualifier":
            condition_dict = items[0]
            condition_dict["conditiontype"] = "where"
            condition_dict["qualifiers"] = items[1]

            return [condition_dict]
         else:
            raise Exception

      # Merging two conditions
      if len(items) == 3:
         merged_list = []
         merged_list += items[0]
         merged_list += items[-1]

         return merged_list

   def filterfullcondition(self, items):
      """
      Defines the full set of qualifiers for a single constraint or all constraints

      Args:
          items ([type]): [description]

      Returns:
          [type]: [description]
      """

      # Only condition, no qualifiers
      if len(items) == 1:
         items[0]["conditiontype"] = "filter"
         return items
      
      # Has potentially a qualifier
      if len(items) == 2:
         if items[1]["type"] == "qualifier":
            condition_dict = items[0]
            condition_dict["conditiontype"] = "filter"
            condition_dict["qualifiers"] = items[1]

            return [condition_dict]
         else:
            raise Exception

      # Merging two conditions
      if len(items) == 3:
         merged_list = []
         merged_list += items[0]
         merged_list += items[-1]

         return merged_list

   
   def querytype(self, items):
      query_dict = {}
      if len(items) == 1:
         query_dict["function"] = None
         query_dict["datatype"] = items[0].data
      else:
         query_dict["function"] = items[0].data
         query_dict["datatype"] = items[1].data

      return query_dict

   def function(self, items):
      return items[0]
   
   def statement(self, items):
      items = [item for item in items if item != "WHERE"]
      items = [item for item in items if item != "QUERY"]

      if len(items) == 1:
         query_dict = {}
         query_dict["querytype"] = items[0]
         query_dict["conditions"] = []

      if len(items) == 2:
         query_dict = {}
         query_dict["querytype"] = items[0]
         query_dict["conditions"] = items[1]
      
      if len(items) == 3:
         query_dict = {}
         query_dict["querytype"] = items[0]
         query_dict["conditions"] = items[1] + items[2]

      return query_dict

   def qualifierfields(self, items):
      return items[0]

   def variable(self, s):
      return s[0].value

   def wildcard(self, s):
      return "ANY"

   # Handling Numerical Expressions
   def plus(self, s):
      return "+"
   def multiply(self, s):
      return "*"
   def minus(self, s):
      return "-"
   def divide(self, s):
      return "/"

   def factor(self, s):
      return s[0]

   def term(self, items):
      if len(items) == 1:
         return items[0]

      has_variable = _has_variable(items)

      string_items = [str(item) for item in items]
      full_expression = "".join(string_items)

      if has_variable:
         return full_expression

      calculated_value = math_parser.parse(full_expression).evaluate({})
      
      return calculated_value

   def numericalexpression(self, items):
      if len(items) == 1:
         return items[0]

      has_variable = _has_variable(items)

      string_items = [str(item) for item in items]
      full_expression = "".join(string_items)

      if has_variable:
         return full_expression

      # Calculating the expression
      calculated_value = math_parser.parse(full_expression).evaluate({})
      
      return calculated_value

   def numericalexpressionwithor(self, items):
      if len(items) == 1:
         return items[0]
            
      # Lets merge these values, based upon type
      merged_list = []

      if isinstance(items[0], list):
         merged_list += items[0]
      else:
         merged_list.append(items[0])
      
      if isinstance(items[-1], list):
         merged_list += items[-1]
      else:
         merged_list.append(items[-1])
         
      return merged_list
    
   def moleculeformula(self, items):
      exact_mass = mass.calculate_mass(formula=items[0])

      return exact_mass

   def aminoacids(self, items):
      exact_mass = mass.calculate_mass(sequence=items[0])
      exact_mass = exact_mass - mass.calculate_mass(formula="H2O")
      return exact_mass

   def peptide(self, items):
      return items[0]

   def peptidecharge(self, items):
      return items[0]

   def peptideion(self, items):
      return items[0]

   def peptidefunction(self, items):
      exact_mass = mass.calculate_mass(sequence=items[0], ion_type=items[2].lower(), charge=int(items[1]))
      return exact_mass
   


   def string(self, s):
      (s,) = s
      return s[1:-1]
      
   def floating(self, n):
      (n,) = n
      return float(n)


def _has_variable(items):
   acceptable_variables = ["X", "Y"]

   for item in items:
      for variable in acceptable_variables:
         if variable in str(item):
            return True

   return False 

def _visualize_parse(input_query, path_to_grammar=None, output_filename="parse.png"):
   if path_to_grammar is None:
      path_to_grammar = os.path.join(os.path.dirname(__file__), "msql.ebnf")

   msql_parser = Lark(open(path_to_grammar).read(), start='statement')
   parsed_tree = msql_parser.parse(input_query)
   tree.pydot__tree_to_png(parsed_tree, output_filename)

def parse_msql(input_query, path_to_grammar=None):
   if path_to_grammar is None:
      path_to_grammar = os.path.join(os.path.dirname(__file__), "msql.ebnf")

   # NOTE: Force capitalization on the input_query, turning this off due to needing lower case in formulas
   # input_query = input_query.upper()

   # Lets try to strip off any comments
   query_splits = input_query.split("\n")
   query_splits = [split.lstrip() for split in query_splits]
   query_splits = [split for split in query_splits if len(split) > 0]
   query_splits = [split.split("#")[0] for split in query_splits]
   query_splits = [split.lstrip() for split in query_splits]
   query_splits = [split for split in query_splits if len(split) > 0]

   input_query = "\n".join(query_splits)

   msql_parser = Lark(open(path_to_grammar).read(), start='statement')
   parsed_tree = msql_parser.parse(input_query)
   parsed_list = MassQLToJSON().transform(parsed_tree)

   parsed_list["query"] = input_query

   return parsed_list



from massql import msql_parser

def translate_query(query, language="english"):
    parsed_query = msql_parser.parse_msql(query)

    sentences = []
    sentences.append(_translate_querytype(parsed_query["querytype"], language=language))

    if len(parsed_query["conditions"]) > 0:
        if language == "english":
            sentences.append("The following conditions are applied to find scans in the mass spec data.")
        if language == "russian":
            sentences.append("Применяются следующие условия для поиска экземпляров в массиве данных.")
        elif language == "korean":
            sentences.append("매스 스펙트럼으로 부터의 스캔 검색조건은 아래와 같습니다.")
        elif language == "chinese":
            sentences.append("这些条件被用于在数据中找到扫描数据。")
        elif language == "french":
            sentences.append("Les conditions suivantes sont appliquées pour trouver les échantillons dans les données du spectrométrie.")
        elif language == "german":
            sentences.append("Die folgenden Bedingungen werden auf massenspektrometrische Daten angewendet um Spektren zu finden.")
        elif language == "spanish":
            sentences.append("Las siguientes condiciones son aplicadas para encontrar los espectros de masas en los datos de espectrometria.")
        elif language == "portuguese":
            sentences.append("As condições a seguir sāo aplicadas para buscar scans nos dados de espectrometria de massas.")
        

    for condition in parsed_query["conditions"]:
        sentences.append(_translate_condition(condition, language=language))

    return "\n".join(sentences)


def _translate_querytype(querytype, language="english"):
    # return information
    ms_level = "MS1"
    if querytype["datatype"] == "datams1data":
        ms_level = "MS1"
    if querytype["datatype"] == "datams2data":
        ms_level = "MS2"

    if querytype["function"] == "functionscaninfo":
        if language == "english":
            return "Returning the scan information on {}.".format(ms_level)
        if language == "russian":
            return "Возвращает информацию о скане на {}.".format(ms_level)
        elif language == "korean":
            return "{} 데이터상의 스캔 정보를 반환합니다.".format(ms_level)
        elif language == "chinese":
            return "返回{}的扫描信息".format(ms_level)
        elif language == "french":
            return "Retourne les informations sur le scan {}.".format(ms_level)
        elif language == "german":
            return "Zurückgeben der Scaninformation von {}.".format(ms_level)
        elif language == "spanish":
            return "Generando información de {}.".format(ms_level)
        elif language == "portuguese":
            return "Encontrando scans com informações de {}.".format(ms_level)
        
    if querytype["function"] == "functionscansum":
        if language == "english":
            return "Returning the summed scan information on {}.".format(ms_level)
        if language == "russian":
            return "Возвращает информацию о суммированном скане на {}.".format(ms_level)
        elif language == "korean":
            return "스칼라를 {} 데이터로 더하기.".format(ms_level)
        elif language == "chinese":
            return "返回{}的总计扫描信息".format(ms_level)
        elif language == "french":
            return "Retourne les informations sur le scan somme de {}.".format(ms_level)
        elif language == "german":
            return "Zurückgeben der zusammengefassten Scaninformation von {}.".format(ms_level)
        elif language == "spanish":
            return "Retornando información de cada espectro de {}.".format(ms_level)
        elif language == "portuguese":
            return "Retornando informação de cada espectro de {}.".format(ms_level)
        


    return "Translator {} not implemented, contact Ming".format(querytype["function"])

def _translate_condition(condition, language="english"):
    if "qualifiers" in condition:
        qualifier_string = " " + _translate_qualifiers(condition["qualifiers"], language=language)
    else:
        qualifier_string = ""

    if condition["type"] == "ms2productcondition":
        if language == "english":
            return "Finding MS2 peak at m/z {}{}.".format(condition["value"][0], qualifier_string) #TODO: add qualifiers
        if language == "russian":
            return "Поиск MS2 пика по m/z {}{}.".format(condition["value"][0], qualifier_string)
        elif language == "korean":
            return "MS2 질량대 전하비 (m/z): {} {}.".format(condition["value"][0], qualifier_string)
        elif language == "chinese":
            return "寻找m/z{}的MS2峰{}".format(condition["value"][0], qualifier_string)
        elif language == "french":
            return "Trouver un pic MS2 à m/z {}{}.".format(condition["value"][0], qualifier_string)
        elif language == "german":
            return "Hat MS2 Signal bei m/z {} {}.".format(condition["value"][0], qualifier_string)
        elif language == "spanish":
            return "Encontrando pico de MS2 a m/z {}{}.".format(condition["value"][0], qualifier_string)
        elif language == "portuguese":
            return "Buscando ions de MS2 de m/z {} {}.".format(condition["value"][0], qualifier_string)

    if condition["type"] == "ms2neutrallosscondition":
        if language == "english":
            return "Finding MS2 neutral loss peak at m/z {}{}.".format(condition["value"][0], qualifier_string) #TODO: add qualifiers
        if language == "russian":
            return "Поиск MS2 нейтраллизации пика по m/z {}{}.".format(condition["value"][0], qualifier_string)
        elif language == "korean":
            return "MS2 전자 소모량 {} {}.".format(condition["value"][0], qualifier_string)
        elif language == "chinese":
            return "在{}的MS2中探测到m/z{}".format(condition["value"][0], qualifier_string)
        elif language == "french":
            return "Trouver un pic MS2 neutral loss à m/z {}{}.".format(condition["value"][0], qualifier_string)
        elif language == "german":
            return "Hat MS2 Neutralverlust von m/z {} {}.".format(condition["value"][0], qualifier_string)
        elif language == "spanish":
            return "Encontrando pico de MS2 de neutral loss en m/z {}{}.".format(condition["value"][0], qualifier_string)
        elif language == "portuguese":
            return "Encontrando peak de neutralização MS2 na m/z {}{}.".format(condition["value"][0], qualifier_string)
    
    if condition["type"] == "ms1mzcondition":
        if language == "english":
            return "Finding MS1 peak at m/z {}{}.".format(condition["value"][0], qualifier_string) #TODO: add qualifiers]
        if language == "russian":
            return "Поиск MS1 пика по m/z {}{}.".format(condition["value"][0], qualifier_string)
        elif language == "korean":
            return "MS1 파워 밀도 {} {}.".format(condition["value"][0], qualifier_string)
        elif language == "chinese":
            return "在{}的MS1中找到m/z{}".format(condition["value"][0], qualifier_string)
        elif language == "french":
            return "Trouver un pic MS1 à m/z {}{}.".format(condition["value"][0], qualifier_string)
        elif language == "german":
            return "Hat MS1 Signal bei m/z {} {}.".format(condition["value"][0], qualifier_string)
        elif language == "spanish":
            return "Encontrando pico de MS1 en m/z {}{}.".format(condition["value"][0], qualifier_string)
        elif language == "portuguese":
            return "Buscando ions de MS1 de m/z {}{}.".format(condition["value"][0], qualifier_string)
    
    if condition["type"] == "ms2precursorcondition":
        if language == "english":
            return "Finding MS2 spectra with a precursor m/z {}{}.".format(condition["value"][0], qualifier_string) #TODO: add qualifiers
        if language == "russian":
            return "Поиск MS2 по предварительному m/z {}{}.".format(condition["value"][0], qualifier_string)
        elif language == "korean":
            return "MS2 파워 밀도 {} {}.".format(condition["value"][0], qualifier_string)
        elif language == "chinese":
            return "在{}的MS2中找到m/z{}".format(condition["value"][0], qualifier_string)
        elif language == "french":
            return "Trouver des spectres MS2 avec une m/z de précursor {}{}.".format(condition["value"][0], qualifier_string)
        elif language == "german":
            return "Hat MS2 Vorläuferion bei m/z {} {}.".format(condition["value"][0], qualifier_string)
        elif language == "spanish":
            return "Encontrando espectros de MS2 con un precursor m/z {}{}.".format(condition["value"][0], qualifier_string)
        elif language == "portuguese":
            return "Buscando espectros MS2 com m/z de referencia {}{}.".format(condition["value"][0], qualifier_string)

    return "Translator {} not implemented, contact Ming".format(condition["type"])

def _translate_qualifiers(qualifiers, language="english"):
    qualifier_phrases = []

    for qualifier in qualifiers:
        # These are keys, so looking them  up
        if "qualifier" in qualifier:
            qualifier_phrases.append(_translate_qualifier(qualifiers[qualifier], language=language))
    
    if language == "english":
        return "and ".join(qualifier_phrases)
    if language == "russian":
        return "и в".join(qualifier_phrases)
    elif language == "korean":
        return "와 ".join(qualifier_phrases)
    elif language == "chinese":
        return "和 ".join(qualifier_phrases)
    elif language == "french":
        return "et ".join(qualifier_phrases)
    elif language == "german":
        return "und ".join(qualifier_phrases)
    elif language == "spanish":
        return "y ".join(qualifier_phrases)
    elif language == "portuguese":
        return "e ".join(qualifier_phrases)

    return " and ".join(qualifier_phrases)

def _translate_qualifier(qualifier, language="english"):
    if qualifier["name"] == "qualifierppmtolerance":
        if language == "english":
            return "a {} PPM tolerance".format(qualifier["value"])
        if language == "russian":
            return "с {} PPM точность".format(qualifier["value"])
        elif language == "korean":
            return "오차범위 (tolerance): {} ppm 의 조건으로 검색합니다".format(qualifier["value"])
        elif language == "chinese":
            return "允许{} ppm质量偏差".format(qualifier["value"])
        elif language == "french":
            return "une tolérance {} PPM".format(qualifier["value"])
        elif language == "german":
            return "eine {} PPM Abweichung".format(qualifier["value"])
        elif language == "spanish":
            return "con tolerancia de {} PPM".format(qualifier["value"])
        elif language == "portuguese":
            return "com {} ppm de tolerância".format(qualifier["value"])

    if qualifier["name"] == "qualifiermztolerance":
        if language == "english":
            return "a {} m/z tolerance".format(qualifier["value"])
        if language == "russian":
            return "с {} m/z точность".format(qualifier["value"])
        elif language == "korean":
            return "m/z 오차 {} 밀도".format(qualifier["value"])
        elif language == "chinese":
            return "一个{} m/z 容差".format(qualifier["value"])
        elif language == "french":
            return "une tolérance {} m/z".format(qualifier["value"])
        elif language == "german":
            return "eine {} m/z Abweichung".format(qualifier["value"])
        elif language == "spanish":
            return "un {} m/z de tolerancia".format(qualifier["value"])
        elif language == "portuguese":
            return "uma tolerância de {} m/z".format(qualifier["value"])

    if qualifier["name"] == "qualifierintensitypercent":
        if language == "english":
            return "a minimum percent intensity relative to base peak of {}%".format(qualifier["value"])
        if language == "russian":
            return "с минимальной процентной интенсивностью {}%".format(qualifier["value"])
        elif language == "korean":
            return "최소 정상 높이 {}%".format(qualifier["value"])
        elif language == "chinese":
            return "一个最低比值{}%的精确度".format(qualifier["value"])
        elif language == "french":
            return "une intensité minimale relative à la base de {}%".format(qualifier["value"])
        elif language == "german":
            return "eine minimale Intensität relative zum Basispeak von {}%".format(qualifier["value"])
        elif language == "spanish":
            return "un mínimo de {}% de intensidad relativa a base pico".format(qualifier["value"])
        elif language == "portuguese":
            return "uma intensidade mínima relativa ao pico base de {}%".format(qualifier["value"])

    if qualifier["name"] == "qualifierintensityreference":
        if language == "english":
            return "this peak is used as the intensity reference for other peaks in the spectrum"
        if language == "russian":
            return "этот пик используется в качестве интенсивности спектра"
        elif language == "korean":
            return "이 데이터를 이용해 시퀀스에 다른 데이터를 인식한다"
        elif language == "chinese":
            return "这个峰用作输入精确度的参考峰"
        elif language == "french":
            return "ce pico est utilisé comme référence d'intensité pour les autres pics de l'échantillon"
        elif language == "german":
            return "dieses Signal wird als Intensitätsreferenz für andere Signale verwendet"
        elif language == "spanish":
            return "este pico é utilizado como referencia de intensidade para outros picos no espectro"
        elif language == "portuguese":
            return "este pico é usado como referencia de intensidade para outros picos no espectro"

    if qualifier["name"] == "qualifierintensitymatch":
        if language == "english":
            return "an expected relative intensity to reference peak of {}".format(qualifier["value"]) #TODO: we should likely remove the Y or assume it 1.0
        if language == "russian":
            return "соответствующая интенсивность {}".format(qualifier["value"])
        elif language == "korean":
            return "예상 비율 이용시 이용할 대상 데이터의 정보 {}".format(qualifier["value"])
        elif language == "chinese":
            return "一个预期相对于参考峰的精确度 {}".format(qualifier["value"])
        elif language == "french":
            return "une intensité prévu relative à la référence de {}".format(qualifier["value"])
        elif language == "german":
            return "eine erwartete Intensität relativ zum Referenzsignal von {}".format(qualifier["value"])
        elif language == "spanish":
            return "una intensidad esperada relativa ao pico de referencia de {}".format(qualifier["value"])
        elif language == "portuguese":
            return "em uma intensidade relativa esperada ao pico de referencia de {}".format(qualifier["value"])

    if qualifier["name"] == "qualifierintensitytolpercent":
        if language == "english":
            return "accepting variability of {}% in relative intensity".format(qualifier["value"])
        if language == "russian":
            return "принимая на случай {}% в относительной интенсивности".format(qualifier["value"])
        elif language == "korean":
            return "이상한 {}%의 비율의 신체 이용 허용".format(qualifier["value"])
        elif language == "chinese":
            return "接受{}%的精确度变动".format(qualifier["value"])
        elif language == "french":
            return "accepter la variabilité de {}% en intensité relative".format(qualifier["value"])
        elif language == "german":
            return "eine Toleranz von {}% in der erwarteten relativen Intensität".format(qualifier["value"])
        elif language == "spanish":
            return "aceptando variabilidad de {}% de intensidad relativa".format(qualifier["value"])
        elif language == "portuguese":
            return "e aceitando uma variabilidade de {}% da intensidade relativa".format(qualifier["value"])

    return "Translator {} not implemented, contact Ming".format(qualifier["name"])



  
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="massql",
    version="0.0.9",
    author="Mingxun Wang",
    author_email="mwang87@gmail.com",
    description="Mass spectrometry query language python implementation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mwang87/MassQueryLanguage",
    project_urls={
        "Bug Tracker": "https://github.com/mwang87/MassQueryLanguage/issues",
        "Documentation": "https://mwang87.github.io/MassQueryLanguage_Documentation/"
    },
    scripts=['massql/msql_cmd.py'],
    entry_points = {
        'console_scripts': ['massql=massql.msql_cmd:main'],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=["massql"],
    install_requires=[
        "pymzml",
        "lark-parser",
        "pandas",
        "pyarrow",
        "ray",
        "tqdm",
        "py_expression_eval",
        "matchms",
        "pyteomics",
        "psims",
        "plotly",
        "kaleido",
        "pydot"
    ],
    python_requires=">=3.6",
    include_package_data=True
)



import json
import os
import pymzml
import pandas as pd
import numpy as np
from tqdm import tqdm
from matchms.importing import load_from_mgf
from pyteomics import mzxml, mzml

import logging
logger = logging.getLogger('msql_fileloading')


def load_data(input_filename, cache=False):
    """
    Loading data generically

    Args:
        input_filename ([type]): [description]
        cache (bool, optional): [description]. Defaults to False.

    Returns:
        [type]: [description]
    """
    if cache:
        ms1_filename = input_filename + "_ms1.msql.feather"
        ms2_filename = input_filename + "_ms2.msql.feather"

        if os.path.exists(ms1_filename) or os.path.exists(ms2_filename):
            try:
                ms1_df = pd.read_feather(ms1_filename)
            except:
                ms1_df = pd.DataFrame()
            try:
                ms2_df = pd.read_feather(ms2_filename)
            except:
                ms2_df = pd.DataFrame()

            return ms1_df, ms2_df

    # Actually loading
    if input_filename[-5:].lower() == ".mzml":
        #ms1_df, ms2_df = _load_data_mzML(input_filename)
        #ms1_df, ms2_df = _load_data_mzML2(input_filename) # Faster version using pymzML
        ms1_df, ms2_df = _load_data_mzML_pyteomics(input_filename) # Faster version using pymzML

    elif input_filename[-6:].lower() == ".mzxml":
        ms1_df, ms2_df = _load_data_mzXML(input_filename)
    
    elif input_filename[-5:] == ".json":
        ms1_df, ms2_df = _load_data_gnps_json(input_filename)
    
    elif input_filename[-4:].lower() == ".mgf":
        ms1_df, ms2_df = _load_data_mgf(input_filename)

    elif input_filename[-4:].lower() == ".txt":
        ms1_df, ms2_df = _load_data_txt(input_filename)
    
    else:
        print("Cannot Load File Extension")
        raise Exception("File Format Not Supported")


    # Saving Cache
    if cache:
        ms1_filename = input_filename + "_ms1.msql.feather"
        ms2_filename = input_filename + "_ms2.msql.feather"

        if not (os.path.exists(ms1_filename) or os.path.exists(ms2_filename)):
            try:
                ms1_df.to_feather(ms1_filename)
            except:
                pass

            try:
                ms2_df.to_feather(ms2_filename)
            except:
                pass

    return ms1_df, ms2_df

def _load_data_mgf(input_filename):
    file = load_from_mgf(input_filename)

    ms2mz_list = []
    for i, spectrum in enumerate(file):
        if len(spectrum.peaks.mz) == 0:
            continue

        mz_list = list(spectrum.peaks.mz)
        i_list = list(spectrum.peaks.intensities)
        i_max = max(i_list)
        i_sum = sum(i_list)

        for i in range(len(mz_list)):
            if i_list[i] == 0:
                continue

            peak_dict = {}
            peak_dict["i"] = i_list[i]
            peak_dict["i_norm"] = i_list[i] / i_max
            peak_dict["i_tic_norm"] = i_list[i] / i_sum
            peak_dict["mz"] = mz_list[i]

            # Handling malformed mgf files
            try:
                peak_dict["scan"] = spectrum.metadata["scans"]
            except:
                peak_dict["scan"] = i + 1
            try:
                peak_dict["rt"] = float(spectrum.metadata["rtinseconds"]) / 60
            except:
                peak_dict["rt"] = 0
            try:
                peak_dict["precmz"] = float(spectrum.metadata["pepmass"][0])
            except:
                peak_dict["precmz"] = 0

            peak_dict["ms1scan"] = 0
            peak_dict["charge"] = 1 # TODO: Add Charge Correctly here
            peak_dict["polarity"] = 1 # TODO: Add Polarity Correctly here

            ms2mz_list.append(peak_dict)

    # Turning into pandas data frames
    ms1_df = pd.DataFrame([peak_dict])
    ms2_df = pd.DataFrame(ms2mz_list)

    return ms1_df, ms2_df

def _load_data_gnps_json(input_filename):
    all_spectra = json.loads(open(input_filename).read())

    ms1_df_list = []
    ms2_df_list = []

    for spectrum in tqdm(all_spectra):
        # Skipping spectra bigger than 1MB of peaks
        if len(spectrum["peaks_json"]) > 1000000:
            continue

        peaks = json.loads(spectrum["peaks_json"])
        peaks = [peak for peak in peaks if peak[1] > 0]
        if len(peaks) == 0:
            continue
        i_max = max([peak[1] for peak in peaks])
        i_sum = sum([peak[1] for peak in peaks])
        if i_max == 0:
            continue

        ms2mz_list = []

        for peak in peaks:
            peak_dict = {}
            peak_dict["i"] = peak[1]
            peak_dict["i_norm"] = peak[1] / i_max
            peak_dict["i_tic_norm"] = peak[1] / i_sum
            peak_dict["mz"] = peak[0]
            peak_dict["scan"] = spectrum["spectrum_id"]
            peak_dict["rt"] = 0
            peak_dict["precmz"] = float(spectrum["Precursor_MZ"])
            peak_dict["ms1scan"] = 0
            peak_dict["charge"] = 1 # TODO: Add Charge Correctly here
            peak_dict["polarity"] = 1 # TODO: Add Polarity Correctly here

            ms2mz_list.append(peak_dict)
        
        # Turning into pandas data frames
        if len(ms2mz_list) > 0:
            ms2_df = pd.DataFrame(ms2mz_list)
            ms2_df_list.append(ms2_df)
            
            ms1_df = pd.DataFrame([peak_dict])
            ms1_df_list.append(ms1_df)

    # Merging
    ms1_df = pd.concat(ms1_df_list).reset_index()
    ms2_df = pd.concat(ms2_df_list).reset_index()

    return ms1_df, ms2_df

def _load_data_mzXML(input_filename):
    ms1mz_list = []
    ms2mz_list = []
    previous_ms1_scan = 0

    with mzxml.read(input_filename) as reader:
        for spectrum in tqdm(reader):
            if len(spectrum["intensity array"]) == 0:
                continue
                
            mz_list = list(spectrum["m/z array"])
            i_list = list(spectrum["intensity array"])
            i_max = max(i_list)
            i_sum = sum(i_list)

            mslevel = spectrum["msLevel"]
            if mslevel == 1:
                for i in range(len(mz_list)):
                    peak_dict = {}
                    peak_dict["i"] = i_list[i]
                    peak_dict["i_norm"] = i_list[i] / i_max
                    peak_dict["i_tic_norm"] = i_list[i] / i_sum
                    peak_dict["mz"] = mz_list[i]
                    peak_dict["scan"] = spectrum["id"]
                    peak_dict["rt"] = spectrum["retentionTime"]
                    peak_dict["polarity"] = _determine_scan_polarity_mzXML(spectrum)

                    ms1mz_list.append(peak_dict)

                    previous_ms1_scan = spectrum["id"]

            if mslevel == 2:
                msn_mz = spectrum["precursorMz"][0]["precursorMz"]
                msn_charge = 0

                if "precursorCharge" in spectrum["precursorMz"][0]:
                    msn_charge = spectrum["precursorMz"][0]["precursorCharge"]
                    
                for i in range(len(mz_list)):
                    peak_dict = {}
                    peak_dict["i"] = i_list[i]
                    peak_dict["i_norm"] = i_list[i] / i_max
                    peak_dict["i_tic_norm"] = i_list[i] / i_sum
                    peak_dict["mz"] = mz_list[i]
                    peak_dict["scan"] = spectrum["id"]
                    peak_dict["rt"] = spectrum["retentionTime"]
                    peak_dict["precmz"] = msn_mz
                    peak_dict["ms1scan"] = previous_ms1_scan
                    peak_dict["charge"] = msn_charge
                    peak_dict["polarity"] = _determine_scan_polarity_mzXML(spectrum)

                    ms2mz_list.append(peak_dict)

    # Turning into pandas data frames
    ms1_df = pd.DataFrame(ms1mz_list)
    ms2_df = pd.DataFrame(ms2mz_list)

    return ms1_df, ms2_df


def _determine_scan_polarity_mzML(spec):
    """
    Gets an enum for positive and negative polarity, for pymzml

    Args:
        spec ([type]): [description]

    Returns:
        [type]: [description]
    """
    polarity = 0
    negative_polarity = spec["negative scan"]
    if negative_polarity is True:
        polarity = 2
    positive_polarity = spec["positive scan"]
    if positive_polarity is True:
        polarity = 1

    return polarity

def _determine_scan_polarity_pyteomics_mzML(spec):
    """
    Gets an enum for positive and negative polarity, for pyteomics

    Args:
        spec ([type]): [description]

    Returns:
        [type]: [description]
    """
    polarity = 0

    if "negative scan" in spec:
        polarity = 2
    if "positive scan" in spec:
        polarity = 1

    return polarity

def _determine_scan_polarity_mzXML(spec):
    polarity = 0
    if spec["polarity"] == "+":
        polarity = 1
    if spec["polarity"] == "-":
        polarity = 2
    return polarity

def _load_data_mzML_pyteomics(input_filename):
    """
    This is a loading operation using pyteomics to help with loading mzML files with ion mobility

    Args:
        input_filename ([type]): [description]
    """

    previous_ms1_scan = 0

    # MS1
    all_mz = []
    all_rt = []
    all_polarity = []
    all_i = []
    all_i_norm = []
    all_i_tic_norm = []
    all_scan = []

    # MS2
    all_msn_mz = []
    all_msn_rt = []
    all_msn_polarity = []
    all_msn_i = []
    all_msn_i_norm = []
    all_msn_i_tic_norm = []
    all_msn_scan = []
    all_msn_precmz = []
    all_msn_ms1scan = []
    all_msn_charge = []
    all_msn_mobility = []

    with mzml.read(input_filename) as reader:
        for spectrum in tqdm(reader):
            if len(spectrum["intensity array"]) == 0:
                continue
            
            # Getting the RT
            try:
                rt = spectrum["scanList"]["scan"][0]["scan start time"]
            except:
                rt = 0
            
            # Correcting the unit
            try:
                if spectrum["scanList"]["scan"][0]["scan start time"].unit_info == "second":
                    rt = rt / 60
            except:
                pass

            scan = int(spectrum["id"].replace("scanId=", "").split("scan=")[-1])
                
            mz = spectrum["m/z array"]
            intensity = spectrum["intensity array"]
            i_max = max(intensity)
            i_sum = sum(intensity)

            mslevel = spectrum["ms level"]
            if mslevel == 1:
                all_mz += list(mz)
                all_i += list(intensity)
                all_i_norm += list(intensity / i_max)
                all_i_tic_norm += list(intensity / i_sum)
                all_rt += len(mz) * [rt]
                all_scan += len(mz) * [scan]
                all_polarity += len(mz) * [_determine_scan_polarity_pyteomics_mzML(spectrum)]

                previous_ms1_scan = scan

            if mslevel == 2:
                msn_mz = spectrum["precursorList"]["precursor"][0]["selectedIonList"]["selectedIon"][0]["selected ion m/z"]
                msn_charge = 0

                if "charge state" in spectrum["precursorList"]["precursor"][0]["selectedIonList"]["selectedIon"][0]:
                    msn_charge = int(spectrum["precursorList"]["precursor"][0]["selectedIonList"]["selectedIon"][0]["charge state"])

                all_msn_mz += list(mz)
                all_msn_i += list(intensity)
                all_msn_i_norm += list(intensity / i_max)
                all_msn_i_tic_norm += list(intensity / i_sum)
                all_msn_rt += len(mz) * [rt]
                all_msn_scan += len(mz) * [scan]
                all_msn_polarity += len(mz) * [_determine_scan_polarity_pyteomics_mzML(spectrum)]
                all_msn_precmz += len(mz) * [msn_mz]
                all_msn_ms1scan += len(mz) * [previous_ms1_scan] 
                all_msn_charge += len(mz) * [msn_charge]

                if "product ion mobility" in spectrum["precursorList"]["precursor"][0]["selectedIonList"]["selectedIon"][0]:
                    mobility = spectrum["precursorList"]["precursor"][0]["selectedIonList"]["selectedIon"][0]["product ion mobility"]
                    all_msn_mobility += len(mz) * [mobility]

    ms1_df = pd.DataFrame()
    if len(all_mz) > 0:
        ms1_df['i'] = all_i
        ms1_df['i_norm'] = all_i_norm
        ms1_df['i_tic_norm'] = all_i_tic_norm
        ms1_df['mz'] = all_mz
        ms1_df['scan'] = all_scan
        ms1_df['rt'] = all_rt
        ms1_df['polarity'] = all_polarity

    ms2_df = pd.DataFrame()
    if len(all_msn_mz) > 0:
        ms2_df['i'] = all_msn_i
        ms2_df['i_norm'] = all_msn_i_norm
        ms2_df['i_tic_norm'] = all_msn_i_tic_norm
        ms2_df['mz'] = all_msn_mz
        ms2_df['scan'] = all_msn_scan
        ms2_df['rt'] = all_msn_rt
        ms2_df["polarity"] = all_msn_polarity
        ms2_df["precmz"] = all_msn_precmz
        ms2_df["ms1scan"] = all_msn_ms1scan
        ms2_df["charge"] = all_msn_charge

        if len(all_msn_mobility) == len(all_msn_i):
            ms2_df["mobility"] = all_msn_mobility
    
    return ms1_df, ms2_df

def _load_data_mzML2(input_filename):
    """This is a faster loading version, but a bit more memory intensive

    Args:
        input_filename ([type]): [description]

    Returns:
        [type]: [description]
    """    


    MS_precisions = {
        1: 5e-6,
        2: 20e-6,
        3: 20e-6,
        4: 20e-6,
        5: 20e-6,
        6: 20e-6,
        7: 20e-6,
    }
    run = pymzml.run.Reader(input_filename, MS_precisions=MS_precisions)

    previous_ms1_scan = 0

    # MS1
    all_mz = []
    all_rt = []
    all_polarity = []
    all_i = []
    all_i_norm = []
    all_i_tic_norm = []
    all_scan = []

    # MS2
    all_msn_mz = []
    all_msn_rt = []
    all_msn_polarity = []
    all_msn_i = []
    all_msn_i_norm = []
    all_msn_i_tic_norm = []
    all_msn_scan = []
    all_msn_precmz = []
    all_msn_ms1scan = []
    all_msn_charge = []

    for i, spec in tqdm(enumerate(run)):
        # Getting RT
        rt = spec.scan_time_in_minutes()

        # Getting peaks
        peaks = spec.peaks("raw")

        # Filtering out zero rows
        peaks = peaks[~np.any(peaks < 1.0, axis=1)]

        if spec.ms_level == 2:
            if len(peaks) > 1000:
                # Sorting by intensity
                peaks = peaks[peaks[:,1].argsort()]

                # Getting top 1000
                peaks = peaks[-1000:]

        if len(peaks) == 0:
            continue
        
        mz, intensity = zip(*peaks)

        i_max = max(intensity)
        i_sum = sum(intensity)

        if spec.ms_level == 1:
            all_mz += list(mz)
            all_i += list(intensity)
            all_i_norm += list(intensity / i_max)
            all_i_tic_norm += list(intensity / i_sum)
            all_rt += len(mz) * [rt]
            all_scan += len(mz) * [spec.ID]
            all_polarity += len(mz) * [_determine_scan_polarity_mzML(spec)]

            previous_ms1_scan = spec.ID

        if spec.ms_level == 2:
            msn_mz = spec.selected_precursors[0]["mz"]
            charge = 0
            if "charge" in spec.selected_precursors[0]:
                charge = spec.selected_precursors[0]["charge"]

            all_msn_mz += list(mz)
            all_msn_i += list(intensity)
            all_msn_i_norm += list(intensity / i_max)
            all_msn_i_tic_norm += list(intensity / i_sum)
            all_msn_rt += len(mz) * [rt]
            all_msn_scan += len(mz) * [spec.ID]
            all_msn_polarity += len(mz) * [_determine_scan_polarity_mzML(spec)]
            all_msn_precmz += len(mz) * [msn_mz]
            all_msn_ms1scan += len(mz) * [previous_ms1_scan] 
            all_msn_charge += len(mz) * [charge]


    ms1_df = pd.DataFrame()
    if len(all_mz) > 0:
        ms1_df['i'] = all_i
        ms1_df['i_norm'] = all_i_norm
        ms1_df['i_tic_norm'] = all_i_tic_norm
        ms1_df['mz'] = all_mz
        ms1_df['scan'] = all_scan
        ms1_df['rt'] = all_rt
        ms1_df['polarity'] = all_polarity

    ms2_df = pd.DataFrame()
    if len(all_msn_mz) > 0:
        ms2_df['i'] = all_msn_i
        ms2_df['i_norm'] = all_msn_i_norm
        ms2_df['i_tic_norm'] = all_msn_i_tic_norm
        ms2_df['mz'] = all_msn_mz
        ms2_df['scan'] = all_msn_scan
        ms2_df['rt'] = all_msn_rt
        ms2_df["polarity"] = all_msn_polarity
        ms2_df["precmz"] = all_msn_precmz
        ms2_df["ms1scan"] = all_msn_ms1scan
        ms2_df["charge"] = all_msn_charge
    
    return ms1_df, ms2_df

def _load_data_mzML(input_filename):
    MS_precisions = {
        1: 5e-6,
        2: 20e-6,
        3: 20e-6,
        4: 20e-6,
        5: 20e-6,
        6: 20e-6,
        7: 20e-6,
    }
    run = pymzml.run.Reader(input_filename, MS_precisions=MS_precisions)

    ms1_df_list = []
    ms2_df_list = []
    previous_ms1_scan = 0

    for i, spec in tqdm(enumerate(run)):
        ms1_df = pd.DataFrame()
        ms2_df = pd.DataFrame()

        # Getting RT
        rt = spec.scan_time_in_minutes()

        # Getting peaks
        peaks = spec.peaks("raw")

        # Filtering out zero rows
        peaks = peaks[~np.any(peaks < 1.0, axis=1)]

        # Sorting by intensity
        peaks = peaks[peaks[:, 1].argsort()]

        if spec.ms_level == 2:
            # Getting top 1000
            peaks = peaks[-1000:]

        if len(peaks) == 0:
            continue

        mz, intensity = zip(*peaks)

        i_max = max(intensity)
        i_sum = sum(intensity)
        
        if spec.ms_level == 1:
            ms1_df['i'] = intensity
            ms1_df['i_norm'] = intensity / i_max
            ms1_df['i_tic_norm'] = intensity / i_sum
            ms1_df['mz'] = mz
            ms1_df['scan'] = spec.ID
            ms1_df['rt'] = rt
            ms1_df['polarity'] = _determine_scan_polarity_mzML(spec)
            
            previous_ms1_scan = spec.ID

        if spec.ms_level == 2:
            msn_mz = spec.selected_precursors[0]["mz"]
            charge = 0
            if "charge" in spec.selected_precursors[0]:
                charge = spec.selected_precursors[0]["charge"]

            ms2_df['i'] = intensity
            ms2_df['i_norm'] = intensity / i_max
            ms2_df['i_tic_norm'] = intensity / i_sum
            ms2_df['mz'] = mz
            ms2_df['scan'] = spec.ID
            ms2_df['rt'] = rt
            ms2_df["polarity"] = _determine_scan_polarity_mzML(spec)
            ms2_df["precmz"] = msn_mz
            ms2_df["ms1scan"] = previous_ms1_scan
            ms2_df["charge"] = charge

        # Turning into pandas data frames
        if len(ms1_df) > 0:
            ms1_df_list.append(ms1_df)
        
        if len(ms2_df) > 0:
            ms2_df_list.append(ms2_df)

    if len(ms1_df_list) > 0:
        ms1_df = pd.concat(ms1_df_list).reset_index()
    else:
        ms1_df = pd.DataFrame()

    if len(ms2_df_list) > 0:
        ms2_df = pd.concat(ms2_df_list).reset_index()
    else:
        ms2_df = pd.DataFrame()

    return ms1_df, ms2_df

def _load_data_txt(input_filename):
    # We are assuming whitespace separated columns, first is mz, second is intensity, and will be marked as MS1
    mz_list = []
    i_list = []
    for line in open(input_filename):
        cleaned_line = line.rstrip()
        if len(cleaned_line) == 0:
            continue
        mz, i = cleaned_line.split()

        mz_list.append(float(mz))
        i_list.append(float(i))
        
    ms1_df = pd.DataFrame()
    ms1_df['mz'] = mz_list
    ms1_df['i'] = i_list
    ms1_df['i_norm'] = ms1_df['i'] / max(ms1_df['i'])
    ms1_df['i_tic_norm'] = ms1_df['i'] / sum(ms1_df['i'])
    ms1_df['scan'] = 1
    ms1_df['rt'] = 0
    ms1_df['polarity'] = "Positive"

    print(ms1_df)

    return ms1_df, pd.DataFrame()



#!/usr/bin/env python

import argparse
import os
import sys
import json
import pandas as pd

# Making sure the root is in the path, kind of a hack
sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))
sys.path.insert(0, os.path.join(os.path.dirname(__file__)))

from massql import msql_parser
from massql import msql_engine
from massql import msql_extract

def main():
    parser = argparse.ArgumentParser(description="MSQL CMD")
    parser.add_argument('filename', help='Input filename')
    parser.add_argument('query', help='Input Query')
    parser.add_argument('--output_file', default=None, help='output results filename')
    parser.add_argument('--parallel_query', default="NO", help='YES to make it parallel with ray locally, NO is default')
    parser.add_argument('--cache', default="YES", help='YES to cache with feather, YES is the default')
    parser.add_argument('--original_path', default=None, help='Original absolute path for the filename, useful in proteosafe')
    parser.add_argument('--extract_mzML', default=None, help='Extracting spectra found as mzML file')
    parser.add_argument('--extract_json', default=None, help='Extracting spectra found as json file')
    
    args = parser.parse_args()

    print(args)

    PARALLEL = args.parallel_query == "YES"

    if PARALLEL:
        msql_engine.init_ray()

    # Massaging the query on input, we have a system to enable multiple queries to be entered, where results are merged
    # The delimeter is specified as |||
    all_queries = args.query.split("|||")

    # Let's parse first
    for query in all_queries:
        parsed_query = msql_parser.parse_msql(query)
        print(json.dumps(parsed_query, indent=4))

    # Executing
    all_results_list = []
    for i, query in enumerate(all_queries):
        results_df = msql_engine.process_query(query, 
                                                args.filename, 
                                                cache=(args.cache == "YES"), 
                                                parallel=PARALLEL)

        results_df["query_index"] = i
        all_results_list.append(results_df)

    # Merging
    results_df = pd.concat(all_results_list)

    print("#############################")
    print("MassQL Found {} results".format(len(results_df)))
    print("#############################")

    # Setting mzupper and mzlower
    try:
        if "comment" in results_df:
            results_df["mz_lower"] = results_df["comment"].astype(float) - 10
            results_df["mz_upper"] = results_df["comment"].astype(float) + 10
    except:
        pass
    
    if args.output_file and len(results_df) > 0:
        results_df["filename"] = os.path.basename(args.filename)

        if args.original_path is not None:
            useful_filename = args.original_path
            # TODO: Clean up for ProteoSAFe
            useful_filename = useful_filename.split("demangled_spectra/")[-1]

            # Cleanup for repository search
            useful_filename = useful_filename.replace("/data/ccms-data/uploads/", "")
            
            # Saving output
            results_df["original_path"] = useful_filename

        # Forcing Types
        try:
            results_df["scan"] = results_df["scan"].astype(int)
        except:
            pass

        try:
            results_df["ms1scan"] = results_df["ms1scan"].astype(int)
        except:
            pass

        try:
            results_df["charge"] = results_df["charge"].astype(int)
        except:
            pass
            
        try:
            results_df["mslevel"] = results_df["mslevel"].astype(int)
        except:
            pass

        # Forcing the column order
        columns = list(results_df.columns)
        columns.sort()

        if ".tsv" in args.output_file:
            results_df.to_csv(args.output_file, index=False, sep="\t", columns=columns)
        else:
            results_df.to_csv(args.output_file, index=False, columns=columns)

        # Extracting
        if args.extract_json is not None and len(results_df) > 0:
            print("Extracting {} spectra".format(len(results_df)))
            try:
                msql_extract._extract_spectra(results_df, os.path.dirname(args.filename), output_json_filename=args.extract_json)
            except:
                print("Extraction Failed")
                

if __name__ == "__main__":
    main()
